"""Calculations of thermal displacements."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings
from typing import Union

import numpy as np

from phonopy.interface.cif import write_cif_P1
from phonopy.phonon.mesh import IterMesh, Mesh
from phonopy.physical_units import get_physical_units


class ThermalMotion:
    """Base class of thermal displacement calculation classes."""

    def __init__(self, iter_mesh: Union[IterMesh, Mesh], freq_min=None, freq_max=None):
        """Init method."""
        self._iter_mesh = iter_mesh
        if freq_min is None:
            self._fmin = 0
        else:
            self._fmin = freq_min
        if freq_max is None:
            self._fmax = None
        else:
            self._fmax = freq_max

        masses = iter_mesh.dynamical_matrix.primitive.masses
        self._masses = masses * get_physical_units().AMU
        self._masses3 = (
            np.array([[m] * 3 for m in masses]).ravel() * get_physical_units().AMU
        )
        self._temperatures = None

    def _get_Q2(self, freq, t):  # freq in THz
        return (
            get_physical_units().Hbar
            * get_physical_units().EV
            / get_physical_units().Angstrom ** 2
            * ((self._get_population(freq, t) + 0.5) / (freq * 1e12 * 2 * np.pi))
        )

    @property
    def temperatures(self):
        """Setter and getter of temperatures."""
        return self._temperatures

    @temperatures.setter
    def temperatures(self, temperatures):
        t_array = np.array(temperatures)
        condition = np.logical_not(t_array < 0)
        self._temperatures = np.extract(condition, t_array)

    def get_temperatures(self):
        """Return temperatures."""
        warnings.warn(
            "ThermalMotion.get_temperatures() is deprecated. "
            "Use temperatures attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.temperatures

    def set_temperatures(self, temperatures):
        """Set temperatures."""
        warnings.warn(
            "ThermalMotion.set_temperatures() is deprecated. "
            "Use temperatures attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.temperatures = temperatures

    def set_temperature_range(self, t_min=None, t_max=None, t_step=None):
        """Set temperatures by range."""
        if t_min is None:
            _t_min = 10
        elif t_min < 0:
            _t_min = 0
        else:
            _t_min = t_min

        if t_max is None:
            _t_max = 1000
        elif t_max > _t_min:
            _t_max = t_max
        else:
            _t_max = _t_min

        if t_step is None:
            _t_step = 10
        elif t_step > 0:
            _t_step = t_step
        else:
            _t_step = 10

        self._temperatures = np.arange(
            _t_min, _t_max + _t_step / 2.0, _t_step, dtype="double"
        )

    def _get_population(self, freq, t):  # freq in THz
        """Return phonon population number.

        Three types of combinations of array inputs are possible.
        - single freq and single t
        - single freq and len(t) > 1
        - len(freq) > 1 and single t

        """
        condition = t > 1.0
        # Avoid using isinstance with bool to distinguish from int.
        if isinstance(condition, (bool, np.bool_)):
            if condition:
                return 1.0 / (
                    np.exp(
                        freq
                        * get_physical_units().THzToEv
                        / (get_physical_units().KB * t)
                    )
                    - 1
                )
            else:
                return 0.0
        else:
            vals = np.zeros(len(t), dtype="double")
            vals[condition] = 1.0 / (
                np.exp(
                    freq
                    * get_physical_units().THzToEv
                    / (get_physical_units().KB * t[condition])
                )
                - 1
            )
            return vals


class ThermalDisplacements(ThermalMotion):
    """Class to calculate thermal displacements (mean square displacements).

    Attributes
    ----------
    thermal_displacements : ndarray
        With `projection_direction` is `None`
            shape=(temps, natoms * 3), dtype=float
        otherwise
            shape=(temps, natoms), dtype=float

    """

    def __init__(
        self,
        iter_mesh: Union[IterMesh, Mesh],
        projection_direction=None,
        freq_min=None,
        freq_max=None,
    ):
        """Init method.

        Parameters
        ----------
        iter_mesh:
            Mesh or IterMesh instance. Grid points must not be reduced by
            symmetry, i.e., IterMesh instance has to be create
            ``is_mesh_symmetry=False``.
        projection_direction:
            Eigenvector projection direction in Cartesian
            coordinates. If None, eigenvector is not projected.
        freq_min:
            Minimum phonon frequency to determine wheather include or not.
        freq_max:
            Maximum phonon frequency to determine wheather include or not.

        """
        super().__init__(iter_mesh, freq_min=freq_min, freq_max=freq_max)
        if projection_direction is None:
            self._projection_direction = None
        else:
            self._projection_direction = projection_direction / np.linalg.norm(
                projection_direction
            )
        self._displacements = None

    @property
    def thermal_displacements(self):
        """Return thermal displacements."""
        return self._displacements

    def get_thermal_displacements(self):
        """Return thermal displacements and temperatures."""
        warnings.warn(
            "ThermalDisplacements.get_thermal_displacements() is deprecated. "
            "Use thermal_displacements and temperatures attributes instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return (self._temperatures, self._displacements)

    def run(self):
        """Calculate thermal displacements."""
        if self._projection_direction is not None:
            masses = self._masses
        else:
            masses = self._masses3
        temps = self._temperatures
        disps = np.zeros((len(temps), len(masses)), dtype=float)

        for count, (fs, vecs) in enumerate(self._iter_mesh):  # noqa B007
            if self._projection_direction is not None:
                p_vecs = np.dot(
                    vecs.T.reshape(-1, 3), self._projection_direction
                ).reshape(-1, len(masses))
                vecs2 = np.abs(p_vecs) ** 2 / masses
            else:
                vecs2 = (abs(vecs) ** 2).T / masses

            valid_indices = fs > self._fmin
            if self._fmax is not None:
                valid_indices *= fs < self._fmax

            if len(temps) == 1:
                Q2 = self._get_Q2(fs[valid_indices], temps[0])
                disps[0] += np.dot(Q2, vecs2[valid_indices])
            else:
                for f, v2 in zip(fs[valid_indices], vecs2[valid_indices]):
                    disps += np.outer(self._get_Q2(f, temps), v2)

        assert np.prod(self._iter_mesh.mesh_numbers) == count + 1
        self._displacements = disps / (count + 1)

    def write_yaml(self, filename="thermal_displacements.yaml"):
        """Write results to file in yaml."""
        natom = len(self._masses)
        lines = []
        lines.append("# Thermal displacements")
        lines.append("natom: %5d" % (natom))
        lines.append("freq_min: %f" % self._fmin)

        lines.append("thermal_displacements:")
        for t, u in zip(self._temperatures, self._displacements):
            lines.append("- temperature:   %15.7f" % t)
            lines.append("  displacements:")
            for i, elems in enumerate(np.reshape(u, (natom, -1))):
                text = "  - [ %10.7f" % elems[0]
                for j in range(len(elems) - 1):
                    text += ", %10.7f" % elems[j + 1]
                text += " ] # atom %d" % (i + 1)
                lines.append(text)

        with open(filename, "w") as w:
            w.write("\n".join(lines))

    def plot(self, pyplot, is_legend=False):
        """Return pyplot of thermal displacements calculation result."""
        xyz = ["x", "y", "z"]
        for i, u in enumerate(self._displacements.transpose()):
            pyplot.plot(
                self._temperatures, u, label=("%d-%s" % (i // 3 + 1, xyz[i % 3]))
            )

        if is_legend:
            pyplot.legend(loc="upper left")

    def _project_eigenvectors(self):
        """Project eigenvectors to specific direction.

        Eigenvectors are projected along Cartesian direction.

        """
        self._p_eigenvectors = []
        for vecs_q in self._eigenvectors:
            p_vecs_q = []
            for vecs in vecs_q.T:
                p_vecs_q.append(np.dot(vecs.reshape(-1, 3), self._projection_direction))
            self._p_eigenvectors.append(np.transpose(p_vecs_q))
        self._p_eigenvectors = np.array(self._p_eigenvectors)


class ThermalDisplacementMatrices(ThermalMotion):
    """Class to calculate thermal displacement (mean square displacement) matrices."""

    def __init__(
        self,
        iter_mesh: Union[IterMesh, Mesh],
        freq_min=None,
        freq_max=None,
        lattice=None,
    ):
        """Init method.

        Parameters
        ----------
        iter_mesh:
            Mesh or IterMesh instance. Grid points must not be reduced by
            symmetry, i.e., IterMesh instance has to be create
            ``is_mesh_symmetry=False``.
        freq_min: float
            Minimum phonon frequency to determine wheather include or not.
        freq_max: float
            Maximum phonon frequency to determine wheather include or not.
        lattice: array_like
            Lattice parameters (column vectors) in real space
            dtype='double', shape=(3, 3)

        """
        super().__init__(iter_mesh, freq_min=freq_min, freq_max=freq_max)
        self._disp_matrices = None
        self._disp_matrices_cif = None

        if lattice is not None:
            A = lattice
            N = np.diag([np.linalg.norm(x) for x in np.linalg.inv(A)])
            self._ANinv = np.linalg.inv(np.dot(A, N))
        else:
            self._ANinv = None

    @property
    def thermal_displacement_matrices(self):
        """Return thermal displacement matrices."""
        return self._disp_matrices

    @property
    def thermal_displacement_matrices_cif(self):
        """Return thermal displacement matrices in cif definition."""
        return self._disp_matrices_cif

    def get_thermal_displacement_matrices(self):
        """Return thermal displacement matrices."""
        warnings.warn(
            "ThermalDisplacementMatrices.get_thermal_displacement_matrices() is "
            "deprecated. Use thermal_displacement_matrices and temperatures "
            "attributes instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return (self._temperatures, self._disp_matrices)

    def run(self, np_overflow=None):
        """Calculate thermal displacement matrices.

        Parameters
        ----------
        np_overflow: str or None
            Switch of error handling of numpy. 'raise' to see which phonon it is.

        """
        np.seterr(over=np_overflow)
        self._get_disp_matrices()
        np.seterr(over=None)

        if self._ANinv is not None:
            self._disp_matrices_cif = np.zeros(
                self._disp_matrices.shape, dtype="double"
            )
            for i, matrices in enumerate(self._disp_matrices):
                for j, mat in enumerate(matrices):
                    mat_cif = np.dot(np.dot(self._ANinv, mat), self._ANinv.T)
                    self._disp_matrices_cif[i, j] = mat_cif

        self._get_disp_matrices()

    def _get_disp_matrices(self):
        dtype_complex = "c%d" % (np.dtype("double").itemsize * 2)
        disps = np.zeros(
            (len(self._temperatures), len(self._masses), 3, 3), dtype=dtype_complex
        )
        for count, (freqs, eigvecs) in enumerate(self._iter_mesh):  # noqa B007
            valid_indices = freqs > self._fmin
            if self._fmax is not None:
                valid_indices *= freqs < self._fmax
            for i_band, (f, vec) in enumerate(
                zip(freqs[valid_indices], (eigvecs.T)[valid_indices])
            ):
                c = np.zeros((len(self._masses), 3, 3), dtype=dtype_complex, order="C")
                for i, (v, m) in enumerate(zip(vec.reshape(-1, 3), self._masses)):
                    c[i] = np.outer(v, v.conj()) / m

                # for i, t in enumerate(self._temperatures):
                #     try:
                #         disps[i] += self._get_Q2(f, t) * np.array(c)
                #     except FloatingPointError as e:
                #         # Probably, overflow in exp(freq / (KB * T))
                #         print("%s: T=%.1f freq=%.2f (band #%d)" %
                #               (e, t, f, i_band))
                try:
                    Q2 = self._get_Q2(f, self._temperatures)
                    disps += Q2[:, None, None, None] * c[None, :, :, :]
                except FloatingPointError as e:
                    # Probably, overflow in exp(freq / (KB * T))
                    print("%s: freq=%.2f (band #%d)" % (e, f, i_band))

        assert np.prod(self._iter_mesh.mesh_numbers) == count + 1
        assert (abs(disps.imag) < 1e-10).all()
        self._disp_matrices = disps.real / (count + 1)

    def write_cif(self, cell, temperature_index, filename="tdispmat.cif"):
        """Write results to file in P1 symmetry CIF format."""
        write_cif_P1(
            cell,
            U_cif=self._disp_matrices_cif[temperature_index],
            filename=filename,
        )

    def write_yaml(self, filename="thermal_displacement_matrices.yaml"):
        """Write results to file in yaml."""
        natom = len(self._masses)
        lines = []

        lines.append("# Thermal displacement_matrices")
        lines.append("natom: %5d" % (natom))
        lines.append("freq_min: %f" % self._fmin)
        lines.append("thermal_displacement_matrices:")
        for i, t in enumerate(self._temperatures):
            matrices = self._disp_matrices[i]
            lines.append("- temperature:   %15.7f" % t)
            lines.append("  displacement_matrices:")
            for j, mat in enumerate(matrices):
                # For checking imaginary part that should be zero
                # lines.append("  - # atom %d" % (i + 1))
                # for v in mat:
                #     lines.append(
                #         "    [ %8.5f, %8.5f, %8.5f, %8.5f, %8.5f, %8.5f ]"
                #         % (tuple(v.real) + tuple(v.imag)))
                m = mat
                lines.append(
                    ("  - [ " + "%8.5f, " * 5 + "%8.5f ] # atom %d")
                    % (m[0, 0], m[1, 1], m[2, 2], m[1, 2], m[0, 2], m[0, 1], j + 1)
                )
            if self._ANinv is not None:
                matrices_cif = self._disp_matrices_cif[i]
                lines.append("  displacement_matrices_cif:")
                for j, mat_cif in enumerate(matrices_cif):
                    m = mat_cif
                    lines.append(
                        ("  - [ " + "%8.5f, " * 5 + "%8.5f ] # atom %d")
                        % (m[0, 0], m[1, 1], m[2, 2], m[1, 2], m[0, 2], m[0, 1], j + 1)
                    )

        with open(filename, "w") as w:
            w.write("\n".join(lines))



"""Show symmetry information invoked by --symmetry command option."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np
import spglib

from phonopy import Phonopy
from phonopy.interface.calculator import (
    get_default_cell_filename,
    write_crystal_structure,
)
from phonopy.interface.phonopy_yaml import PhonopyYaml
from phonopy.interface.vasp import sort_positions_by_symbols
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import determinant, guess_primitive_matrix
from phonopy.structure.symmetry import Symmetry


def check_symmetry(phonon: Phonopy, cell_info: dict):
    """Check symmetry of input cell.

    This function first standardizes the input crystal structure based on
    symmetry and determines the primitive cell and unit cell. If
    `supercell_matrix` is not provided, it displays the symmetry information and
    writes it to files. If `supercell_matrix` is specified, the function
    generates a file named `phonopy_supercell.yaml`, which contains details
    about the primitive cell, unit cell, and supercell.

    """
    base_fname = get_default_cell_filename(phonon.calculator)
    symprec = phonon.primitive_symmetry.tolerance
    (bravais_lattice, bravais_pos, bravais_numbers) = spglib.refine_cell(
        phonon.primitive.totuple(), symprec
    )
    _, _, _, perm = sort_positions_by_symbols(bravais_numbers)
    bravais = PhonopyAtoms(
        numbers=bravais_numbers[perm],
        scaled_positions=bravais_pos[perm],
        cell=bravais_lattice,
    )
    trans_mat = guess_primitive_matrix(bravais, symprec=symprec)
    ph = Phonopy(
        bravais,
        supercell_matrix=cell_info["supercell_matrix"],
        primitive_matrix=trans_mat,
        calculator=phonon.calculator,
    )

    if (
        cell_info["phonopy_yaml"] is None
        or cell_info["phonopy_yaml"].supercell_matrix is None
    ) and determinant(cell_info["supercell_matrix"]) > 1:
        print(f'Input crystal structure: "{cell_info["optional_structure_info"][0]}"')
        phyml = PhonopyYaml().set_phonon_info(ph)
        phyml.frequency_unit_conversion_factor = None
        phyml.symmetry = None
        with open("phonopy_supercell.yaml", "w") as w:
            print(phyml, file=w)

        print(
            "\n".join(
                [
                    "Standardized primitive cell, unit cell, and supercell have been "
                    "written in ",
                    '"phonopy_supercell.yaml". This file can be read by "phonopy-load" '
                    "command.",
                ]
            )
        )
    else:
        _show_symmetry_yaml(phonon, cell_info, base_fname, ph)


def _show_symmetry_yaml(phonon: Phonopy, cell_info: dict, base_fname: str, ph: Phonopy):
    print(
        _get_symmetry_yaml(phonon.primitive, phonon.primitive_symmetry, phonon.version)
    )
    print(f"# Input crystal structure: \"{cell_info['optional_structure_info'][0]}'")
    if cell_info["phonopy_yaml"] is None:
        optional_structure_info = cell_info["optional_structure_info"]
        filename = "B" + base_fname
        print(f'# Symmetrized conventional unit cell is written into "{filename}" and')
        write_crystal_structure(
            filename,
            ph.unitcell,
            interface_mode=phonon.calculator,
            optional_structure_info=optional_structure_info,
        )
        filename = "P" + base_fname
        print(f'# Symmetrized primitive is written into "{filename}" and ')
        write_crystal_structure(
            filename,
            ph.primitive,
            interface_mode=phonon.calculator,
            optional_structure_info=optional_structure_info,
        )
    _write_symcells_yaml(ph)


def _write_symcells_yaml(ph: Phonopy):
    phyml = PhonopyYaml().set_phonon_info(ph)
    phyml.supercell_matrix = None
    phyml.supercell = None
    phyml.frequency_unit_conversion_factor = None
    phyml.symmetry = None
    with open("phonopy_symcells.yaml", "w") as w:
        print(phyml, file=w)

    print(
        '# Unit cell and primitive cell have been written in "phonopy_symcells.yaml".'
    )
    print("# These structures can be read in a python script as follows:")
    print("#")
    print("# import phonopy")
    print('# ph = phonopy.load("phonopy_symcells.yaml")')
    print("# unitcell = ph.unitcell")
    print("# primitive_cell = ph.primitive")


def _get_symmetry_yaml(cell: PhonopyAtoms, symmetry: Symmetry, phonopy_version=None):
    rotations = symmetry.symmetry_operations["rotations"]
    translations = symmetry.symmetry_operations["translations"]

    atom_sets = symmetry.get_map_atoms()
    independent_atoms = symmetry.get_independent_atoms()
    wyckoffs = symmetry.get_Wyckoff_letters()

    lines = []

    if phonopy_version is not None:
        lines.append("phonopy_version: '%s'" % phonopy_version)

    if cell.magnetic_moments is None:
        spg_symbol, spg_number = symmetry.get_international_table().split()
        spg_number = int(spg_number.replace("(", "").replace(")", ""))
        lines.append("space_group_type: '%s'" % spg_symbol)
        lines.append("space_group_number: %d" % spg_number)
        lines.append("point_group_type: '%s'" % symmetry.pointgroup_symbol)
    lines.append("space_group_operations:")
    for i, (r, t) in enumerate(zip(rotations, translations)):
        lines.append("- rotation: # %d" % (i + 1))
        for vec in r:
            lines.append("  - [%2d, %2d ,%2d]" % tuple(vec))
        line = "  translation: ["
        for j, x in enumerate(t):
            if abs(x - np.rint(x)) < 1e-5:
                line += " 0.00000"
            else:
                line += "%8.5f" % x
            if j < 2:
                line += ", "
            else:
                line += " ]"
                lines.append(line)
    lines.append("atom_mapping:")
    for i, atom_num in enumerate(atom_sets):
        lines.append("  %d: %d" % (i + 1, atom_num + 1))
    lines.append("site_symmetries:")
    for i in independent_atoms:
        sitesym = symmetry.get_site_symmetry(i)
        lines.append("- atom: %d" % (i + 1))

        if cell.magnetic_moments is None:
            lines.append("  Wyckoff: '%s'" % wyckoffs[i])
        site_pointgroup = spglib.get_pointgroup(sitesym)
        lines.append("  site_point_group: '%s'" % site_pointgroup[0].strip())
        lines.append("  orientation:")
        for v in site_pointgroup[2]:
            lines.append("  - [%2d, %2d, %2d]" % tuple(v))

        lines.append("  rotations:")
        for j, r in enumerate(sitesym):
            lines.append("  - # %d" % (j + 1))
            for vec in r:
                lines.append("    - [%2d, %2d, %2d]" % tuple(vec))

    return "\n".join(lines)



"""Core routines for QHA."""

# Copyright (C) 2012 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np

from phonopy.physical_units import get_physical_units
from phonopy.qha.eos import fit_to_eos, get_eos


class BulkModulus:
    """Bulk modulus class.

    This class is used to calculate bulk modulus only from temperature
    independent energy input.

    """

    def __init__(self, volumes, energies, pressure=None, eos="vinet"):
        """Init method.

        volumes : array_like
            Unit cell volumes where energies are obtained.
            shape=(volumes, ), dtype='double'.
        energies : array_like
            Energies obtained at volumes and optionally temperatures.
            shape=(volumes, ), or shape=(temperatures, volumes) dtype='double'.
        pressure: float, optional
            Pressure in GPa that is added to energy as PV term.
        eos : str, optional
            Identifier of equation of states function.

        """
        self._volumes = np.array(volumes)
        self._energies = np.array(energies)
        self._eos_name = eos
        self._eos = get_eos(self._eos_name)

        if pressure is not None:
            self._energies += (
                self._volumes * pressure / get_physical_units().EVAngstromToGPa
            )

        if self._energies.ndim == 1:
            self._energy = None
            self._bulk_modulus = None
            self._b_prime = None
            self._equiv_volume = None
            (
                self._energy,
                self._bulk_modulus,
                self._b_prime,
                self._equiv_volume,
            ) = self.fit_to_eos(self._energies)
        elif self._energies.ndim == 2:
            self._energy = np.zeros(len(self._energies), dtype="double")
            self._bulk_modulus = np.zeros_like(self._energy)
            self._b_prime = np.zeros_like(self._energy)
            self._equiv_volume = np.zeros_like(self._energy)
            for i, energies_at_T in enumerate(self._energies):
                e, b, bp, ev = self.fit_to_eos(energies_at_T)
                self._energy[i] = e
                self._bulk_modulus[i] = b
                self._b_prime[i] = bp
                self._equiv_volume[i] = ev
        else:
            raise TypeError("Array shape of energies is wrong.")

    def fit_to_eos(self, energies):
        """Fit energy-volume to EOS."""
        try:
            e, b, bp, ev = fit_to_eos(self._volumes, energies, self._eos)
        except TypeError as exc:
            msg = ['Failed to fit to "%s" equation of states.' % self._eos_name]
            if len(self._volumes) < 4:
                msg += ["At least 4 volume points are needed for the fitting."]
            msg += ["Careful choice of volume points is recommended."]
            raise RuntimeError("\n".join(msg)) from exc
        return e, b, bp, ev

    @property
    def bulk_modulus(self):
        """Return bulk modulus."""
        return self._bulk_modulus

    def get_bulk_modulus(self):
        """Return bulk modulus."""
        warnings.warn(
            "BulkModulus.get_bulk_modulus() is deprecated."
            "Use BulkModulus.bulk_modulus attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.bulk_modulus

    @property
    def equilibrium_volume(self):
        """Return volume at equilibrium."""
        return self._equiv_volume

    def get_equilibrium_volume(self):
        """Return volume at equilibrium."""
        warnings.warn(
            "BulkModulus.get_equilibrium_volume() is deprecated."
            "Use BulkModulus.equilibrium_volume attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.equilibrium_volume

    @property
    def b_prime(self):
        """Return fitted parameter B'."""
        return self._b_prime

    def get_b_prime(self):
        """Return fitted parameter B'."""
        warnings.warn(
            "BulkModulus.get_b_prime() is deprecated."
            "Use BulkModulus.b_prime attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._b_prime

    @property
    def energy(self):
        """Return fitted parameter of energy."""
        return self._energy

    def get_energy(self):
        """Return fitted parameter of energy."""
        warnings.warn(
            "BulkModulus.get_energy() is deprecated.Use BulkModulus.energy attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._energy

    def get_parameters(self):
        """Return fitted parameters."""
        return (self._energy, self._bulk_modulus, self._b_prime, self._equiv_volume)

    def get_eos(self):
        """Return EOS function as a python method."""
        warnings.warn(
            "BulkModulus.get_eos() is deprecated.", DeprecationWarning, stacklevel=2
        )
        return self._eos

    def plot(self, thin_number=10):
        """Plot fitted EOS curve."""
        import matplotlib.pyplot as plt

        vols = self._volumes
        volume_points = np.linspace(min(vols), max(vols), 201)
        _, ax = plt.subplots()
        parameters = self.get_parameters()
        if self._energies.ndim == 1:
            ax.plot(volume_points, self._eos(volume_points, *parameters), "r-")
            ax.plot(vols, self._energies, "bo", markersize=4)
        elif self._energies.ndim == 2:
            for i, (e_t, b_t, bp_t, ev_t) in enumerate(zip(*parameters)):
                if i % thin_number == 0:
                    ep = (e_t, b_t, bp_t, ev_t)
                    ax.plot(volume_points, self._eos(volume_points, *ep), "-")
                    ax.plot(vols, self._energies[i], "bo", markersize=4)
        return plt


class QHA:
    """Quasi harmonic approximation class."""

    def __init__(
        self,
        volumes,  # angstrom^3
        electronic_energies,  # eV
        temperatures,  # K
        cv,  # J/K/mol
        entropy,  # J/K/mol
        fe_phonon,  # kJ/mol
        pressure=None,
        eos="vinet",
        t_max=None,
        energy_plot_factor=None,
    ):
        """Init method.

        Parameters
        ----------
        volumes: array_like
            Unit cell volumes (V) in angstrom^3.
            dtype='double'
            shape=(volumes,)
        electronic_energies: array_like
            Electronic energies (U_el) or electronic free energies (F_el) in eV.
            It is assumed as formar if ndim==1 and latter if ndim==2.
            dtype='double'
            shape=(volumes,) or (temperatuers, volumes)
        temperatures: array_like
            Temperatures ascending order (T) in K.
            dtype='double'
            shape=(temperatures,)
        cv: array_like
            Phonon Heat capacity at constant volume in J/K/mol.
            dtype='double'
            shape=(temperatuers, volumes)
        entropy: array_like
            Phonon entropy at constant volume (S_ph) in J/K/mol.
            dtype='double'
            shape=(temperatuers, volumes)
        fe_phonon: array_like
            Phonon Helmholtz free energy (F_ph) in kJ/mol.
            dtype='double'
            shape=(temperatuers, volumes)
        pressure: float, optional
            Pressure in GPa that is added to energy as PV term.
        eos: str, optional
            Equation of state used for fitting F vs V.
            'vinet', 'murnaghan' or 'birch_murnaghan'.
        t_max: float
            Maximum temperature to be calculated. This has to be not
            greater than the temperature of the third element from the
            end of 'temperatre' elements. If max_t=None, the temperature
            of the third element from the end is used.
        energy_plot_factor: float
            This value is multiplied to energy like values only in plotting.

        """
        self._volumes = np.array(volumes)
        self._electronic_energies = np.array(electronic_energies)
        if pressure is not None:
            self._electronic_energies += (
                self._volumes * pressure / get_physical_units().EVAngstromToGPa
            )
        self._all_temperatures = np.array(temperatures)
        self._cv = np.array(cv)
        self._entropy = np.array(entropy)
        self._fe_phonon = np.array(fe_phonon) / get_physical_units().EvTokJmol

        self._eos = get_eos(eos)
        self._t_max = t_max
        self._energy_plot_factor = energy_plot_factor

        self._temperatures = None
        self._equiv_volumes = None
        self._equiv_energies = None
        self._equiv_bulk_modulus = None
        self._equiv_parameters = None
        self._free_energies = None
        self._num_elems = None

        self._thermal_expansions = None
        self._cp_numerical = None
        self._volume_entropy_parameters = None
        self._volume_cv_parameters = None
        self._volume_entropy = None
        self._volume_cv = None
        self._cp_polyfit = None
        self._dsdv = None
        self._gruneisen_parameters = None
        self._len = None

    @property
    def thermal_expansion(self):
        """Return volumetric thermal expansion coefficients at temperatures."""
        return self._thermal_expansions[: self._len]

    @property
    def helmholtz_volume(self):
        """Return Helmholtz free energies at temperatures and volumes."""
        return self._free_energies[: self._len]

    @property
    def volume_temperature(self):
        """Return equilibrium volumes at temperatures."""
        return self._equiv_volumes[: self._len]

    @property
    def gibbs_temperature(self):
        """Return Gibbs free energies at temperatures."""
        return self._equiv_energies[: self._len]

    @property
    def bulk_modulus_temperature(self):
        """Return bulk modulus vs temperature data."""
        return self._equiv_bulk_modulus[: self._len]

    @property
    def heat_capacity_P_numerical(self):
        """Return heat capacities at constant pressure at temperatures.

        Values are computed by numerical derivative of Gibbs free energy.

        """
        return self._cp_numerical[: self._len]

    @property
    def heat_capacity_P_polyfit(self):
        """Return heat capacities at constant pressure at temperatures.

        Volumes are computed in another way to heat_capacity_P_numerical
        for the better numerical behaviour. But this does not work
        when temperature dependent electronic_energies is supplied.

        """
        if self._electronic_energies.ndim == 1:
            return self._cp_polyfit[: self._len]
        else:
            raise NotImplementedError(
                "QHA.heat_capacity_P_polyfit() unavailable with electronic free energy"
            )

    @property
    def gruneisen_temperature(self):
        """Return Gruneisen parameters at temperatures."""
        return self._gruneisen_parameters[: self._len]

    def run(self, verbose=False):
        """Fit parameters to EOS at temperatures.

        Even if fitting failed, simply omit the volume point. In this case,
        the failed temperature point doesn't exist in the returned arrays.

        """
        if verbose:
            print(("#%11s" + "%14s" * 4) % ("T", "E_0", "B_0", "B'_0", "V_0"))

        # Plus one temperature point is necessary for computing e.g. beta.
        num_elems = self._get_num_elems(self._all_temperatures) + 1
        if num_elems > len(self._all_temperatures):
            num_elems -= 1

        temperatures = []
        parameters = []
        free_energies = []

        for i in range(num_elems):  # loop over temperaturs
            if self._electronic_energies.ndim == 1:
                el_energy = self._electronic_energies
            else:
                el_energy = self._electronic_energies[i]
            fe = [ph_e + el_e for ph_e, el_e in zip(self._fe_phonon[i], el_energy)]

            try:
                ep = fit_to_eos(self._volumes, fe, self._eos)
            except TypeError:
                print("Fitting failure at T=%.1f" % self._all_temperatures[i])

            if ep is None:
                # Simply omit volume point where the fitting failed.
                continue
            else:
                [ee, eb, ebp, ev] = ep
                t = self._all_temperatures[i]
                temperatures.append(t)
                parameters.append(ep)
                free_energies.append(fe)

                if verbose:
                    print(
                        ("%14.6f" * 5)
                        % (
                            t,
                            ep[0],
                            ep[1] * get_physical_units().EVAngstromToGPa,
                            ep[2],
                            ep[3],
                        )
                    )

        self._free_energies = np.array(free_energies)
        self._temperatures = np.array(temperatures)
        self._equiv_parameters = np.array(parameters)
        self._equiv_volumes = np.array(self._equiv_parameters[:, 3])
        self._equiv_energies = np.array(self._equiv_parameters[:, 0])
        self._equiv_bulk_modulus = np.array(
            self._equiv_parameters[:, 1] * get_physical_units().EVAngstromToGPa
        )

        self._num_elems = len(self._temperatures)

        # For computing following values at temperatures, finite difference
        # method is used. Therefore number of temperature points are needed
        # larger than self._num_elems that nearly equals to the temparature
        # point we expect.
        self._set_thermal_expansion()
        self._set_heat_capacity_P_numerical()
        self._set_heat_capacity_P_polyfit()
        self._set_gruneisen_parameter()  # To be run after thermal expansion.

        self._len = len(self._thermal_expansions)
        assert self._len + 1 == self._num_elems

    def plot(self, thin_number=10, volume_temp_exp=None):
        """Plot three figures.

        - Helmholtz free energy at volumes and temperatures.
        - Equilibrium volumes at temperatures.
        - Thermal expansion coefficients at temperatures.

        """
        import matplotlib.pyplot as plt

        plt.rcParams["pdf.fonttype"] = 42
        plt.rcParams["font.family"] = "serif"

        fig, axs = plt.subplots(1, 3, figsize=(7, 3.5))
        axs[0].xaxis.set_ticks_position("both")
        axs[0].yaxis.set_ticks_position("both")
        axs[0].xaxis.set_tick_params(which="both", direction="in")
        axs[0].yaxis.set_tick_params(which="both", direction="in")
        self._plot_helmholtz_volume(axs[0], thin_number=thin_number)
        axs[1].xaxis.set_ticks_position("both")
        axs[1].yaxis.set_ticks_position("both")
        axs[1].xaxis.set_tick_params(which="both", direction="in")
        axs[1].yaxis.set_tick_params(which="both", direction="in")
        self._plot_volume_temperature(axs[1], exp_data=volume_temp_exp)
        axs[2].xaxis.set_ticks_position("both")
        axs[2].yaxis.set_ticks_position("both")
        axs[2].xaxis.set_tick_params(which="both", direction="in")
        axs[2].yaxis.set_tick_params(which="both", direction="in")
        self._plot_thermal_expansion(axs[2])
        plt.tight_layout()
        return plt

    def get_helmholtz_volume(self):
        """Return Helmholtz free energy at volumes and temperatures."""
        warnings.warn(
            "QHA.get_helmholtz_volume() is deprecated.Use helmholtz_volume attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.helmholtz_volume

    def plot_helmholtz_volume(
        self, thin_number=10, xlabel=r"Volume $(\AA^3)$", ylabel="Free energy"
    ):
        """Return pyplot of Helmholtz free energes vs volume at temperatures."""
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        self._plot_helmholtz_volume(
            ax, thin_number=thin_number, xlabel=xlabel, ylabel=ylabel
        )
        return plt

    def plot_pdf_helmholtz_volume(
        self, thin_number=10, filename="helmholtz-volume.pdf"
    ):
        """Plot Helmholtz free energes vs volume at temperatures in pdf."""
        import matplotlib.pyplot as plt

        self._set_rcParams(plt)

        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")

        self._plot_helmholtz_volume(ax, thin_number=thin_number)
        plt.savefig(filename)
        plt.close()

    def write_helmholtz_volume(self, filename="helmholtz-volume.dat"):
        """Write Helmholtz free energy vs volume in file."""
        w = open(filename, "w")
        for i, (t, ep, fe) in enumerate(
            zip(self._temperatures, self._equiv_parameters, self._free_energies)
        ):
            if i == self._len:
                break

            w.write("# Temperature: %f\n" % t)
            w.write("# Parameters: %f %f %f %f\n" % tuple(ep))
            for j, v in enumerate(self._volumes):
                w.write("%20.15f %25.15f\n" % (v, fe[j]))
            w.write("\n\n")
        w.close()

    def write_helmholtz_volume_fitted(
        self, thin_number, filename="helholtz-volume_fitted.dat"
    ):
        """Write Helmholtz free energy (fitted) vs volume in file."""
        if self._energy_plot_factor is None:
            _energy_plot_factor = 1
        else:
            _energy_plot_factor = self._energy_plot_factor

        volume_points = np.linspace(min(self._volumes), max(self._volumes), 201)
        selected_volumes = []
        selected_energies = []

        for i, _ in enumerate(self._temperatures[: self._len]):
            if i % thin_number == 0:
                selected_volumes.append(self._equiv_volumes[i])
                selected_energies.append(self._equiv_energies[i])

        e0 = 0
        for i, t in enumerate(self._temperatures[: self._len]):
            if t >= 298:
                if i > 0:
                    de = self._equiv_energies[i] - self._equiv_energies[i - 1]
                    dt = t - self._temperatures[i - 1]
                    e0 = (
                        298 - self._temperatures[i - 1]
                    ) / dt * de + self._equiv_energies[i - 1]
                else:
                    e0 = 0
                break
        e0 *= _energy_plot_factor
        _data_vol_points = []
        _data_eos = []
        for i, _ in enumerate(self._temperatures[: self._len]):
            if i % thin_number == 0:
                _data_vol_points.append(
                    np.array(self._free_energies[i]) * _energy_plot_factor - e0
                )
                _data_eos.append(
                    self._eos(volume_points, *self._equiv_parameters[i])
                    * _energy_plot_factor
                    - e0
                )

        data_eos = np.array(_data_eos).T
        data_vol_points = np.array(_data_vol_points).T
        data_min = np.array(selected_energies) * _energy_plot_factor - e0

        with open(filename, "w") as w:
            w.write("# Volume points\n")
            for j, k in zip(self._volumes, data_vol_points):
                w.write("%10.5f " % j)
                for ll in k:
                    w.write("%10.5f" % ll)
                w.write("\n")
            w.write("\n# Fitted data\n")

            for m, n in zip(volume_points, data_eos):
                w.write("%10.5f " % m)
                for ll in n:
                    w.write("%10.5f" % ll)
                w.write("\n")
            w.write("\n# Minimas\n")
            for a, b in zip(selected_volumes, data_min):
                w.write("%10.5f %10.5f %s" % (a, b, "\n"))
            w.write("\n")

    def get_volume_temperature(self):
        """Return equilibrium volumes at temperatures."""
        warnings.warn(
            "QHA.get_volume_temperature() is deprecated."
            "Use volume_temperature attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.volume_temperature

    def plot_volume_temperature(self, exp_data=None):
        """Return pyplot of volume vs temperature."""
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        self._plot_volume_temperature(ax, exp_data=exp_data)

        return plt

    def plot_pdf_volume_temperature(
        self, exp_data=None, filename="volume-temperature.pdf"
    ):
        """Plot volume vs temperature in pdf."""
        import matplotlib.pyplot as plt

        self._set_rcParams(plt)

        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")

        self._plot_volume_temperature(ax, exp_data=exp_data)
        plt.savefig(filename)
        plt.close()

    def write_volume_temperature(self, filename="volume-temperature.dat"):
        """Write volume vs temperature in file."""
        w = open(filename, "w")
        for i in range(self._len):
            w.write(
                "%25.15f %25.15f\n" % (self._temperatures[i], self._equiv_volumes[i])
            )
        w.close()

    def get_thermal_expansion(self):
        """Return thermal expansion coefficients at temperatures."""
        warnings.warn(
            "QHA.get_thermal_expansion() is deprecated."
            "Use thermal_expansion attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.thermal_expansion

    def plot_thermal_expansion(self):
        """Return pyplot of thermal expansion vs temperature."""
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        self._plot_thermal_expansion(ax)

        return plt

    def plot_pdf_thermal_expansion(self, filename="thermal_expansion.pdf"):
        """Plot thermal expansion vs temperature in pdf."""
        import matplotlib.pyplot as plt

        self._set_rcParams(plt)

        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")

        self._plot_thermal_expansion(ax)
        plt.savefig(filename)
        plt.close()

    def write_thermal_expansion(self, filename="thermal_expansion.dat"):
        """Write thermal expansion vs temperature in file."""
        w = open(filename, "w")
        for i in range(self._len):
            w.write(
                "%25.15f %25.15f\n"
                % (self._temperatures[i], self._thermal_expansions[i])
            )
        w.close()

    def get_gibbs_temperature(self):
        """Return Gibbs free energies at temperatures."""
        return self.gibbs_temperature

    def plot_gibbs_temperature(
        self, xlabel="Temperature (K)", ylabel="Gibbs free energy"
    ):
        """Return pyplot Gibbs free energy vs temperature."""
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        self._plot_gibbs_temperature(ax, xlabel=xlabel, ylabel=ylabel)

        return plt

    def plot_pdf_gibbs_temperature(self, filename="gibbs-temperature.pdf"):
        """Plot Gibbs free energy vs temperature in pdf."""
        import matplotlib.pyplot as plt

        self._set_rcParams(plt)

        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")

        self._plot_gibbs_temperature(ax)
        plt.savefig(filename)
        plt.close()

    def write_gibbs_temperature(self, filename="gibbs-temperature.dat"):
        """Write Gibbs free energy vs temperature in file."""
        w = open(filename, "w")
        for i in range(self._len):
            w.write(
                "%20.15f %25.15f\n" % (self._temperatures[i], self._equiv_energies[i])
            )
        w.close()

    def get_bulk_modulus_temperature(self):
        """Return bulk moduli at temperatures."""
        return self.bulk_modulus_temperature

    def plot_bulk_modulus_temperature(
        self, xlabel="Temperature (K)", ylabel="Bulk modulus"
    ):
        """Return pyplot of bulk modulus vs temperature."""
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        self._plot_bulk_modulus_temperature(ax, xlabel=xlabel, ylabel=ylabel)
        return plt

    def plot_pdf_bulk_modulus_temperature(
        self, filename="bulk_modulus-temperature.pdf"
    ):
        """Plot bulk modulus vs temperature in pdf."""
        import matplotlib.pyplot as plt

        self._set_rcParams(plt)

        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")

        self._plot_bulk_modulus_temperature(ax)
        plt.savefig(filename)
        plt.close()

    def write_bulk_modulus_temperature(self, filename="bulk_modulus-temperature.dat"):
        """Write bulk modulus vs temperature in file."""
        w = open(filename, "w")
        for i in range(self._len):
            w.write(
                "%20.15f %25.15f\n"
                % (self._temperatures[i], self._equiv_bulk_modulus[i])
            )
        w.close()

    def get_heat_capacity_P_numerical(self):
        """Return C_P by numerical differenciation at temperatures."""
        return self.heat_capacity_P_numerical

    def plot_heat_capacity_P_numerical(self, Z=1, exp_data=None):
        """Return pyplot of C_P by numerical difference vs temperature."""
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        self._plot_heat_capacity_P_numerical(ax, Z=Z, exp_data=exp_data)

        return plt

    def plot_pdf_heat_capacity_P_numerical(
        self, exp_data=None, filename="Cp-temperature.pdf"
    ):
        """Plot C_P by numerical difference vs temperature in pdf."""
        import matplotlib.pyplot as plt

        self._set_rcParams(plt)

        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")

        self._plot_heat_capacity_P_numerical(ax, exp_data=exp_data)
        plt.savefig(filename)
        plt.close()

    def write_heat_capacity_P_numerical(self, filename="Cp-temperature.dat"):
        """Write C_P by numerical difference vs temperature in file."""
        w = open(filename, "w")
        for i in range(self._len):
            w.write(
                "%20.15f %20.15f\n" % (self._temperatures[i], self._cp_numerical[i])
            )
        w.close()

    def get_heat_capacity_P_polyfit(self):
        """Return C_P by fitting at temperatures."""
        return self.heat_capacity_P_polyfit

    def plot_heat_capacity_P_polyfit(self, Z=1, exp_data=None):
        """Return pyplot of C_P by fittings vs temperature."""
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        self._plot_heat_capacity_P_polyfit(ax, Z=Z, exp_data=exp_data)

        return plt

    def plot_pdf_heat_capacity_P_polyfit(
        self, exp_data=None, filename="Cp-temperature_polyfit.pdf"
    ):
        """Plot C_P by fittings vs temperature in pdf."""
        import matplotlib.pyplot as plt

        self._set_rcParams(plt)

        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")

        self._plot_heat_capacity_P_polyfit(ax, exp_data=exp_data)
        plt.savefig(filename)
        plt.close()

    def write_heat_capacity_P_polyfit(
        self,
        filename="Cp-temperature_polyfit.dat",
        filename_ev="entropy-volume.dat",
        filename_cvv="Cv-volume.dat",
        filename_dsdvt="dsdv-temperature.dat",
    ):
        """Write C_P by fittings vs temperature in file."""
        wve = open(filename_ev, "w")
        wvcv = open(filename_cvv, "w")
        for i in range(1, self._len):
            t = self._temperatures[i]
            wve.write("# temperature %20.15f\n" % t)
            wve.write(
                "# %20.15f %20.15f %20.15f %20.15f %20.15f\n"
                % tuple(self._volume_entropy_parameters[i - 1])
            )
            wvcv.write("# temperature %20.15f\n" % t)
            wvcv.write(
                "# %20.15f %20.15f %20.15f %20.15f %20.15f\n"
                % tuple(self._volume_cv_parameters[i - 1])
            )
            for ve, vcv in zip(self._volume_entropy[i - 1], self._volume_cv[i - 1]):
                wve.write("%20.15f %20.15f\n" % tuple(ve))
                wvcv.write("%20.15f %20.15f\n" % tuple(vcv))
            wve.write("\n\n")
            wvcv.write("\n\n")
        wve.close()
        wvcv.close()

        w = open(filename, "w")
        for i in range(self._len):
            w.write("%20.15f %20.15f\n" % (self._temperatures[i], self._cp_polyfit[i]))
        w.close()

        w = open(filename_dsdvt, "w")  # GPa
        for i in range(self._len):
            w.write(
                "%20.15f %20.15f\n"
                % (
                    self._temperatures[i],
                    self._dsdv[i] * 1e21 / get_physical_units().Avogadro,
                )
            )
        w.close()

    def get_gruneisen_temperature(self):
        """Return Grueneisen parameters at temperatures."""
        return self.gruneisen_temperature

    def plot_gruneisen_temperature(self):
        """Return pyplot of Grueneisen parameter vs temperature."""
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        self._plot_gruneisen_temperature(ax)

        return plt

    def plot_pdf_gruneisen_temperature(self, filename="gruneisen-temperature.pdf"):
        """Plot Grueneisen parameter vs temperature in pdf."""
        import matplotlib.pyplot as plt

        self._set_rcParams(plt)

        fig, ax = plt.subplots()
        ax.xaxis.set_ticks_position("both")
        ax.yaxis.set_ticks_position("both")
        ax.xaxis.set_tick_params(which="both", direction="in")
        ax.yaxis.set_tick_params(which="both", direction="in")

        self._plot_gruneisen_temperature(ax)
        plt.savefig(filename)
        plt.close()

    def write_gruneisen_temperature(self, filename="gruneisen-temperature.dat"):
        """Write Grueneisen parameter vs temperature in file."""
        w = open(filename, "w")
        for i in range(self._len):
            w.write(
                "%20.15f %25.15f\n"
                % (self._temperatures[i], self._gruneisen_parameters[i])
            )
        w.close()

    def _plot_helmholtz_volume(
        self, ax, thin_number=10, xlabel=r"Volume $(\AA^3)$", ylabel="Free energy"
    ):
        if self._energy_plot_factor is None:
            _energy_plot_factor = 1
            _ylabel = ylabel + " (eV)"
        else:
            _energy_plot_factor = self._energy_plot_factor
            _ylabel = ylabel

        volume_points = np.linspace(min(self._volumes), max(self._volumes), 201)
        selected_volumes = []
        selected_energies = []

        thin_index = 0
        for i, _ in enumerate(self._temperatures[: self._len]):
            if i % thin_number == 0:
                selected_volumes.append(self._equiv_volumes[i])
                selected_energies.append(self._equiv_energies[i])

        e0 = 0
        for i, t in enumerate(self._temperatures[: self._len]):
            if t >= 298:
                if i > 0:
                    de = self._equiv_energies[i] - self._equiv_energies[i - 1]
                    dt = t - self._temperatures[i - 1]
                    e0 = (
                        298 - self._temperatures[i - 1]
                    ) / dt * de + self._equiv_energies[i - 1]
                else:
                    e0 = 0
                break
        e0 *= _energy_plot_factor

        for i, _ in enumerate(self._temperatures[: self._len]):
            if i % thin_number == 0:
                ax.plot(
                    self._volumes,
                    np.array(self._free_energies[i]) * _energy_plot_factor - e0,
                    "bo",
                    markeredgecolor="b",
                    markersize=3,
                )
                ax.plot(
                    volume_points,
                    self._eos(volume_points, *self._equiv_parameters[i])
                    * _energy_plot_factor
                    - e0,
                    "b-",
                )
                thin_index = i

        for i, j in enumerate((0, thin_index)):
            ax.text(
                self._volumes[-2],
                (self._free_energies[j, -1] + (1 - i * 2) * 0.1 - 0.05)
                * _energy_plot_factor
                - e0,
                "%dK" % int(self._temperatures[j]),
                fontsize=8,
            )

        ax.plot(
            selected_volumes,
            np.array(selected_energies) * _energy_plot_factor - e0,
            "ro-",
            markeredgecolor="r",
            markersize=3,
        )
        ax.set_xlabel(xlabel)
        ax.set_ylabel(_ylabel)

    def _plot_volume_temperature(
        self, ax, exp_data=None, xlabel="Temperature (K)", ylabel=r"Volume $(\AA^3)$"
    ):
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.plot(self._temperatures[: self._len], self._equiv_volumes[: self._len], "r-")
        ax.set_xlim(self._temperatures[0], self._temperatures[self._len - 1])
        # exp
        if exp_data:
            ax.plot(exp_data[0], exp_data[1], "ro")

    def _plot_thermal_expansion(
        self,
        ax,
        xlabel="Temperature (K)",
        ylabel=r"Thermal expansion $(\mathrm{K}^{-1})$",
    ):
        from matplotlib.ticker import ScalarFormatter

        class FixedScaledFormatter(ScalarFormatter):
            def __init__(self):
                super().__init__(useMathText=True)

            def _set_orderOfMagnitude(self, range):
                self.orderOfMagnitude = -6

        ax.yaxis.set_major_formatter(FixedScaledFormatter())
        ax.ticklabel_format(style="sci", axis="y", scilimits=(0, 0))

        beta = np.array(self._thermal_expansions)
        ax.plot(self._temperatures[: self._len], beta[: self._len], "r-")
        ax.set_xlim(self._temperatures[0], self._temperatures[self._len - 1])
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

    def _plot_gibbs_temperature(
        self, ax, xlabel="Temperature (K)", ylabel="Gibbs free energy (eV)"
    ):
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.plot(
            self._temperatures[: self._len], self._equiv_energies[: self._len], "r-"
        )
        ax.set_xlim(self._temperatures[0], self._temperatures[self._len - 1])

    def _plot_bulk_modulus_temperature(
        self, ax, xlabel="Temperature (K)", ylabel="Bulk modulus (GPa)"
    ):
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.plot(
            self._temperatures[: self._len], self._equiv_bulk_modulus[: self._len], "r-"
        )
        ax.set_xlim(self._temperatures[0], self._temperatures[self._len - 1])

    def _plot_heat_capacity_P_numerical(
        self,
        ax,
        Z=1,
        exp_data=None,
        xlabel="Temperature (K)",
        ylabel=r"$C\mathrm{_P}$ $\mathrm{(J/mol\cdot K)}$",
    ):
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.plot(
            self._temperatures[: self._len],
            np.array(self._cp_numerical[: self._len]) / Z,
            "r-",
        )

        # exp
        if exp_data:
            ax.plot(exp_data[0], exp_data[1], "ro")

        ax.set_xlim(self._temperatures[0], self._temperatures[self._len - 1])

    def _plot_heat_capacity_P_polyfit(
        self,
        ax,
        Z=1,
        exp_data=None,
        xlabel="Temperature (K)",
        ylabel=r"$C\mathrm{_P}$ $\mathrm{(J/mol\cdot K)}$",
    ):
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.plot(
            self._temperatures[: self._len],
            np.array(self._cp_polyfit[: self._len]) / Z,
            "r-",
        )

        # exp
        if exp_data:
            ax.plot(exp_data[0], exp_data[1], "ro")

        ax.set_xlim(self._temperatures[0], self._temperatures[self._len - 1])

    def _plot_gruneisen_temperature(
        self, ax, xlabel="Temperature (K)", ylabel="Gruneisen parameter"
    ):
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.plot(
            self._temperatures[: self._len],
            self._gruneisen_parameters[: self._len],
            "r-",
        )
        ax.set_xlim(self._temperatures[0], self._temperatures[self._len - 1])

    def _set_thermal_expansion(self):
        beta = [0.0]
        for i in range(1, self._num_elems - 1):
            dt = self._temperatures[i + 1] - self._temperatures[i - 1]
            dv = self._equiv_volumes[i + 1] - self._equiv_volumes[i - 1]
            beta.append(dv / dt / self._equiv_volumes[i])

        self._thermal_expansions = beta

    def _set_heat_capacity_P_numerical(self):
        cp = []
        g = np.array(self._equiv_energies) * get_physical_units().EvTokJmol * 1000
        cp.append(0.0)

        for i in range(1, self._num_elems - 1):
            t = self._temperatures[i]
            parameters = np.polyfit(
                self._temperatures[i - 1 : i + 2], g[i - 1 : i + 2], 2
            )
            cp.append(-(2 * parameters[0]) * t)

        self._cp_numerical = cp

    def _set_heat_capacity_P_polyfit(self):
        cp = [0.0]
        dsdv = [0.0]
        self._volume_entropy_parameters = []
        self._volume_cv_parameters = []
        self._volume_entropy = []
        self._volume_cv = []

        for j in range(1, self._num_elems - 1):
            t = self._temperatures[j]
            x = self._equiv_volumes[j]

            try:
                parameters = np.polyfit(self._volumes, self._cv[j], 4)
            except np.lib.polynomial.RankWarning as exc:
                msg = ["Failed to fit heat capacities to polynomial of degree 4."]
                if len(self._volumes) < 5:
                    msg += ["At least 5 volume points are needed for the fitting."]
                raise RuntimeError("\n".join(msg)) from exc

            cv_p = np.dot(parameters, np.array([x**4, x**3, x**2, x, 1]))
            self._volume_cv_parameters.append(parameters)

            try:
                parameters = np.polyfit(self._volumes, self._entropy[j], 4)
            except np.lib.polynomial.RankWarning as exc:
                msg = ["Failed to fit entropies to polynomial of degree 4."]
                if len(self._volumes) < 5:
                    msg += ["At least 5 volume points are needed for the fitting."]
                raise RuntimeError("\n".join(msg)) from exc

            dsdv_t = np.dot(parameters[:4], np.array([4 * x**3, 3 * x**2, 2 * x, 1]))
            self._volume_entropy_parameters.append(parameters)

            try:
                parameters = np.polyfit(
                    self._temperatures[j - 1 : j + 2],
                    self._equiv_volumes[j - 1 : j + 2],
                    2,
                )
            except np.lib.polynomial.RankWarning as exc:
                msg = (
                    "Failed to fit equilibrium volumes vs T to polynomial of degree 2."
                )
                raise RuntimeError(msg) from exc
            dvdt = parameters[0] * 2 * t + parameters[1]

            cp.append(cv_p + t * dvdt * dsdv_t)
            dsdv.append(dsdv_t)

            self._volume_cv.append(np.array([self._volumes, self._cv[j]]).T)
            self._volume_entropy.append(np.array([self._volumes, self._entropy[j]]).T)

        self._cp_polyfit = cp
        self._dsdv = dsdv

    def _set_gruneisen_parameter(self):
        gamma = [0]
        for i in range(1, self._num_elems - 1):
            v = self._equiv_volumes[i]
            kt = self._equiv_bulk_modulus[i]
            beta = self._thermal_expansions[i]
            try:
                parameters = np.polyfit(self._volumes, self._cv[i], 4)
            except np.lib.polynomial.RankWarning as exc:
                msg = ["Failed to fit heat capacities to polynomial of degree 4."]
                if len(self._volumes) < 5:
                    msg += ["At least 5 volume points are needed for the fitting."]
                raise RuntimeError("\n".join(msg)) from exc
            cv = (
                np.dot(parameters, [v**4, v**3, v**2, v, 1])
                / v
                / 1000
                / get_physical_units().EvTokJmol
                * get_physical_units().EVAngstromToGPa
            )
            if cv < 1e-10:
                gamma.append(0.0)
            else:
                gamma.append(beta * kt / cv)
        self._gruneisen_parameters = gamma

    def _get_num_elems(self, temperatures):
        if self._t_max is None:
            return len(temperatures)
        else:
            i = np.argmin(np.abs(temperatures - self._t_max))
            return i + 1

    def _set_rcParams(self, plt):
        plt.rcParams["backend"] = "PDF"
        plt.rcParams["pdf.fonttype"] = 42
        plt.rcParams["font.family"] = "serif"
        plt.rcParams["axes.labelsize"] = 18
        plt.rcParams["figure.subplot.left"] = 0.25
        plt.rcParams["figure.subplot.bottom"] = 0.15
        plt.rcParams["figure.figsize"] = 4, 6



"""Routines to collect crystal structure information."""

# Copyright (C) 2018 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

from typing import Optional

import numpy as np

from phonopy.file_IO import is_file_phonopy_yaml
from phonopy.interface.calculator import (
    get_default_cell_filename,
    read_crystal_structure,
)
from phonopy.interface.phonopy_yaml import PhonopyYaml
from phonopy.interface.vasp import read_vasp


def collect_cell_info(
    supercell_matrix=None,
    primitive_matrix=None,
    interface_mode=None,
    cell_filename=None,
    chemical_symbols=None,
    enforce_primitive_matrix_auto=False,
    phonopy_yaml_cls: Optional[type[PhonopyYaml]] = None,
    load_phonopy_yaml: bool = False,
) -> dict:
    """Collect crystal structure information from input file and parameters.

    This function returns crystal structure information obtained from an input
    file and parameters. Although this function is convenient, this function
    tends to be complicated and error-prone since phonopy has to support various
    file formats.

    How crystal structure is recognized
    -----------------------------------
    Phonopy supports various crystal structure file formats. Most of them are
    those from force calculators (e.g., VASP, QE). To let phonopy know which
    calculator format is chosen, usually the calculator interface name is
    simultaneously given. Otherwise, it is considered as the default interface,
    i.e., VASP-like format. When a calculator interface is specified, phonopy
    goes into the calculator interface mode. In this calculator interface mode,
    when none of crystal structure is provided, the default file name for each
    calculator interface mode is assumed and the file name is searched in the
    current directory. The crystal structure information found in this way is
    recognized as the unit cell. The supercell and primitive matrices
    information are given by the parameters of this function.

    When srystal structure search explained above failed, phonopy.yaml like file
    is searched in the current directory. phonopy.yaml like file name can be
    specified as the input crystal structure. Since phonopy.yaml like file
    contains supercell and primitive cell matricies information, these
    parameter inputs of this function are ignored.

    Parameters
    ----------
    supercell_matrix : array_like or None
        3x3 transformation matrix or when it is a diagonal matrix, three
        diagonal elements. Default is None. See also shape_supercell_matrix.
    primitive_matrix : array_like, str, or None
        3x3 transformation matrix or a character representing centring or None.
        Default is None. See also get_primitive_matrix.
    interface_mode : str or None
        Force calculator or crystal structure format name.
    cell_filename : str or None
        Input cell filename.
    chemical_symbols : list of str
        List of chemical symbols or unit cell.
    enforce_primitive_matrix_auto : bool
        Enforce primitive_matrix='auto' when True. Default is False.
    phonopy_yaml_cls : Class object, optional
        PhonopyYaml like class name. This is used to return its instance when
        needed. Default is None, which means PhonopyYaml class type. This can be
        Phono3pyYaml class type. Default is None.
    load_phonopy_yaml : bool
        True means phonopy-load mode.

    Returns
    -------
    dict :
        "unitcell": PhonopyAtoms
            Unit cell.
        "supercell_matrix": ndarray
        "primitive_matrix": ndarray
        "optional_structure_info": list
            See read_crystal_structure.
        "interface_mode": str
            Force calculator or crystal structure format name.
        "phonopy_yaml": None or instance of the class given by phonopy_yaml_cls
            Not None when crystal structure was read phonopy.yaml like file.
        "error_message" : str
            Unless error exists, this entry should not be in this dict.
            Otherwise, some error exists. The error message is storedin the
            string.

    """
    # In some cases, interface mode falls back to phonopy_yaml mode.
    if load_phonopy_yaml:
        fallback_reason = "load_phonopy_yaml mode"
    else:
        fallback_reason = _fallback_to_phonopy_yaml(
            supercell_matrix,
            interface_mode,
            cell_filename,
        )

    _cell_filename = cell_filename
    if load_phonopy_yaml or fallback_reason:
        _interface_mode = "phonopy_yaml"
        keyword = phonopy_yaml_cls.command_name
        if _cell_filename is None:
            pass
        elif is_file_phonopy_yaml(_cell_filename, yaml_dict_keys=[keyword]):
            # Readable as a phonopy.yaml
            pass
        elif is_file_phonopy_yaml(_cell_filename):
            # Readable as yaml
            pass
        else:
            # Not readable as yaml. Proceed to look for default file names.
            _cell_filename = None
    elif interface_mode is None:
        _interface_mode = None
    else:
        _interface_mode = interface_mode.lower()

    unitcell, optional_structure_info = read_crystal_structure(
        filename=_cell_filename,
        interface_mode=_interface_mode,
        chemical_symbols=chemical_symbols,
        phonopy_yaml_cls=phonopy_yaml_cls,
    )

    # Error check
    if unitcell is None:
        err_msg = _get_error_message(
            optional_structure_info,
            fallback_reason,
            cell_filename,  # original cell_filename can be needed for error message.
            phonopy_yaml_cls,
        )
        return {"error_message": err_msg}

    # Retrieve more information on cells
    (
        interface_mode_out,
        supercell_matrix_out,
        primitive_matrix_out,
    ) = _collect_cells_info(
        _interface_mode,
        optional_structure_info,
        interface_mode,
        supercell_matrix,
        primitive_matrix,
    )

    err_msg = []
    unitcell_filename = optional_structure_info[0]
    if supercell_matrix_out is None:
        if _interface_mode == "phonopy_yaml":
            err_msg.append(f"'supercell_matrix' not found  in \"{unitcell_filename}\".")
        else:
            err_msg.append("Supercell matrix information (DIM or --dim) was not found.")
        if _cell_filename is None and (
            unitcell_filename == get_default_cell_filename(interface_mode_out)
        ):
            err_msg += [
                "",
                "Phonopy read the crystal structure from the file having the default "
                "filename ",
                "of each calculator. In this case, supercell matrix has to be "
                "specified.",
                "Because this is the old style way of using %s,"
                % phonopy_yaml_cls.command_name,
            ]
            filenames = [
                '"%s"' % name for name in phonopy_yaml_cls.default_filenames[:-1]
            ]
            err_msg += [
                '"%s" was read being prefered to files such as ' % unitcell_filename,
                '%s, or "%s".'
                % (", ".join(filenames), phonopy_yaml_cls.default_filenames[-1]),
            ]
            err_msg += [
                "",
                'If crystal structure is expected to be read from some "*.yaml" file,',
                'Please rename "%s" to something else.' % unitcell_filename,
            ]
    if np.linalg.det(unitcell.cell) < 0.0:
        err_msg.append("Lattice vectors have to follow the right-hand rule.")
    if len(err_msg) > 0:
        err_msg = [
            'Crystal structure was read from "%s".' % unitcell_filename
        ] + err_msg
        return {"error_message": "\n".join(err_msg)}

    if enforce_primitive_matrix_auto:
        primitive_matrix_out = "auto"

    if _interface_mode == "phonopy_yaml":
        phpy_yaml: PhonopyYaml = optional_structure_info[1]
    else:
        phpy_yaml = None

    return {
        "unitcell": unitcell,
        "supercell_matrix": supercell_matrix_out,
        "primitive_matrix": primitive_matrix_out,
        "optional_structure_info": optional_structure_info,
        "interface_mode": interface_mode_out,
        "phonopy_yaml": phpy_yaml,
    }


def _fallback_to_phonopy_yaml(supercell_matrix, interface_mode, cell_filename):
    """Find possibility to fallback to phonopy.yaml mode.

    Fallback happens in any of the following cases.

    1. Parsing crystal structure file in the VASP POSCAR-style failed
    2. Default VASP POSCAR-style file is not found.
    3. supercell_matrix is not given along with (1) or (2).

    Parameters
    ----------
    supercell_matrix : array_like or None
        None is given when phonopy.yaml mode is expected.
    interface_mode : str or None
        None is the default mode, i.e., VASP like.
    cell_filename : str or None
        File name of VASP style crystal structure. None means the default
        file name, "POSCAR".

    Returns
    -------
    fallback_reason : str or None
        This provides information how to handle after the fallback.
        None means fallback to phonopy.yaml mode will not happen.

    """
    fallback_reason = None

    if interface_mode is None:
        fallback_reason = _poscar_failed(cell_filename)

    if fallback_reason is not None:
        if supercell_matrix is None:
            fallback_reason = "no supercell matrix given"

    return fallback_reason


def _poscar_failed(cell_filename):
    """Determine if fall back happens.

    1) read_vasp (parsing POSCAR-style file) is failed. --> fallback

    ValueError is raised by read_vasp when the POSCAR-style format
    is broken. By this way, we assume the input crystal structure
    is not in the POSCAR-style format and is in the phonopy.yaml
    type.

    2) The file given by get_default_cell_filename('vasp') is not
       found at the current directory.  --> fallback

    This is the trigger to look for the phonopy.yaml type file.

    3) The given file with cell_filename is not found.  --> not fallback

    This case will not invoke phonopy.yaml mode and here nothing
    is done, i.e., fallback_reason = None.
    This error will be caught in the following part again be
    handled properly (read_crystal_structure).

    """
    fallback_reason = None
    try:
        if cell_filename is None:
            read_vasp(get_default_cell_filename("vasp"))
        else:
            read_vasp(cell_filename)
    except ValueError:
        # (1) see above
        fallback_reason = "read_vasp parsing failed"
    except FileNotFoundError:
        if cell_filename is None:
            # (2) see above
            fallback_reason = "default file not found"
        else:
            # (3) see above
            pass
    return fallback_reason


def _collect_cells_info(
    _interface_mode,
    optional_structure_info,
    interface_mode,
    supercell_matrix,
    primitive_matrix,
):
    if _interface_mode == "phonopy_yaml" and optional_structure_info[1] is not None:
        phpy: PhonopyYaml = optional_structure_info[1]
        if phpy.calculator is None:
            interface_mode_out = interface_mode
        else:
            interface_mode_out = phpy.calculator
        if phpy.supercell_matrix is None:
            _supercell_matrix = supercell_matrix
        else:
            _supercell_matrix = phpy.supercell_matrix
        if primitive_matrix is not None:
            _primitive_matrix = primitive_matrix
        elif phpy.primitive_matrix is not None:
            _primitive_matrix = phpy.primitive_matrix
        else:
            _primitive_matrix = None
    else:
        interface_mode_out = _interface_mode
        _supercell_matrix = supercell_matrix
        _primitive_matrix = primitive_matrix

    supercell_matrix_out = _supercell_matrix
    primitive_matrix_out = _primitive_matrix

    return interface_mode_out, supercell_matrix_out, primitive_matrix_out


def _get_error_message(
    optional_structure_info,
    fallback_reason,
    cell_filename,
    phonopy_yaml_cls: type[PhonopyYaml],
):
    """Show error message for failure of getting crystal structure."""
    final_cell_filename = optional_structure_info[0]

    # No fallback to phonopy_yaml mode.
    if fallback_reason is None:
        msg_list = []
        if cell_filename is None:
            msg_list += [
                "Crystal structure file was not specified.",
                "Tried to find default crystal structure file.",
            ]
        msg_list.append(
            f'Crystal structure file "{final_cell_filename}" was not found.'
        )
        return "\n".join(msg_list)

    ####################################
    # Must be phonopy_yaml mode below. #
    ####################################

    msg_list = []
    if fallback_reason in ["default file not found", "read_vasp parsing failed"]:
        if cell_filename:
            vasp_filename = cell_filename
        else:
            vasp_filename = get_default_cell_filename("vasp")

        if fallback_reason == "read_vasp parsing failed":
            msg_list += [
                f'Parsing crystal structure file "{vasp_filename}" '
                "as in VASP format failed.",
                "(Calculator option is needed for parsing different crystal "
                "structure format.)",
            ]
        else:
            msg_list.append(f'Crystal structure file "{vasp_filename}" was not found.')

    elif fallback_reason == "no supercell matrix given":
        msg_list.append("Supercell matrix (DIM or --dim) was not explicitly specified.")

    msg_list.append(f"Switched on {phonopy_yaml_cls.command_name}-yaml mode.")

    if final_cell_filename is None:  # No phonopy*.yaml file was found.
        filenames = [f'"{name}"' for name in phonopy_yaml_cls.default_filenames]
        if len(filenames) == 1:
            text = filenames[0]
        elif len(filenames) == 2:
            text = " or ".join(filenames)
        else:
            tail = " or ".join(filenames[-2:])
            head = ", ".join(filenames[:-2])
            text = head + ", " + tail
        msg_list.append(f"{text} could not be found.")
        return "\n".join(msg_list)

    phpy = optional_structure_info[1]
    if phpy is None:  # Failed to parse phonopy*.yaml.
        msg_list.append(f'But parsing "{final_cell_filename}" failed.')

    return "\n".join(msg_list)



"""Collection of physical units."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings
from math import pi, sqrt

from phonopy.physical_units import get_physical_units

warnings.warn(
    "phonopy.units.* is deprecated. Use phonopy.physical_units.physical_units instead.",
    DeprecationWarning,
    stacklevel=2,
)

kb_J = get_physical_units().KB_J
PlanckConstant = get_physical_units().PlanckConstant
Avogadro = get_physical_units().Avogadro
SpeedOfLight = get_physical_units().SpeedOfLight
AMU = get_physical_units().AMU
EV = get_physical_units().EV
Me = get_physical_units().Me
THz = get_physical_units().THz
Angstrom = get_physical_units().Angstrom
Newton = get_physical_units().Newton
Joule = get_physical_units().Joule
Hbar = get_physical_units().Hbar
Epsilon0 = get_physical_units().Epsilon0
Bohr = get_physical_units().Bohr
Hartree = get_physical_units().Hartree
Rydberg = get_physical_units().Rydberg
Kb = get_physical_units().KB
THzToCm = get_physical_units().THzToCm
THzToEv = get_physical_units().THzToEv
CmToEv = get_physical_units().CmToEv
EVAngstromToGPa = get_physical_units().EVAngstromToGPa
EvTokJmol = get_physical_units().EvTokJmol

VaspToTHz = get_physical_units().DefaultToTHz  # [THz] 15.633302
Wien2kToTHz = (
    sqrt(Rydberg / 1000 * EV / AMU) / (Bohr * 1e-10) / (2 * pi) / 1e12
)  # [THz] 3.44595837
AbinitToTHz = sqrt(EV / (AMU * Bohr)) / Angstrom / (2 * pi) / 1e12  # [THz] 21.49068
PwscfToTHz = (
    sqrt(Rydberg * EV / AMU) / (Bohr * 1e-10) / (2 * pi) / 1e12
)  # [THz] 108.97077
ElkToTHz = (
    sqrt(Hartree * EV / AMU) / (Bohr * 1e-10) / (2 * pi) / 1e12
)  # [THz] 154.10794
SiestaToTHz = sqrt(EV / (AMU * Bohr)) / Angstrom / (2 * pi) / 1e12  # [THz] 21.49068
CP2KToTHz = (
    sqrt(Hartree * EV / (AMU * Bohr)) / Angstrom / (2 * pi) / 1e12
)  # CP2K uses a.u. for forces but Angstrom for distances
CrystalToTHz = VaspToTHz
CastepToTHz = VaspToTHz
DftbpToTHz = (
    sqrt(Hartree * EV / AMU) / (Bohr * 1e-10) / (2 * pi) / 1e12
)  # [THz] 154.10794344
TurbomoleToTHz = ElkToTHz  # Turbomole uses atomic units (Hartree/Bohr)
FleurToTHz = ElkToTHz  # Fleur uses atomic units (Hartree/Bohr)



"""Collection of physical units.

Use get_physical_units() for getting the physical units.
To overwrite the physical units, use set_physical_units().

"""

# Copyright (C) 2025 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from dataclasses import asdict, dataclass
from math import pi, sqrt
from typing import Optional


@dataclass
class PhysicalUnitsGenerator:
    """Physical units for phonon calculations."""

    KB_J: float  # [J/K]
    PlanckConstant: float  # [eV s]
    Avogadro: float
    SpeedOfLight: float  # [m/s]
    AMU: float  # [kg]
    EV: float  # [J]
    Me: float  # [kg]
    THz: Optional[float] = None  # [/s]
    Angstrom: Optional[float] = None  # [m]
    Newton: Optional[float] = None  # [kg m / s^2]
    Joule: Optional[float] = None  # [kg m^2 / s^2]
    Hbar: Optional[float] = None  # [eV s]
    Epsilon0: Optional[float] = None  # [C^2 / N m^2]
    Bohr: Optional[float] = None  # Bohr radius [A]
    Hartree: Optional[float] = None  # Hartree [eV]
    Rydberg: Optional[float] = None  # Rydberg [eV]
    THzToEv: Optional[float] = None  # [eV]
    KB: Optional[float] = None  # [eV/K]
    THzToCm: Optional[float] = None  # [cm^-1]
    CmToEv: Optional[float] = None  # [eV]
    EVAngstromToGPa: Optional[float] = None
    EvTokJmol: Optional[float] = None  # [kJ/mol]
    DefaultToTHz: Optional[float] = None  # [THz]

    def __post_init__(self):
        """Initialize derived physical constants."""
        Mu0 = 4.0e-7 * pi  # [Hartree/m]
        self.THz = 1.0e12  # [/s]
        self.Angstrom = 1.0e-10  # [m]
        self.Newton = 1.0  # [kg m / s^2]
        self.Joule = 1.0  # [kg m^2 / s^2]
        self.Hbar = self.PlanckConstant / (2 * pi)  # [eV s]
        self.Epsilon0 = 1.0 / Mu0 / self.SpeedOfLight**2  # [C^2 / N m^2]
        self.Bohr = (
            4e10 * pi * self.Epsilon0 * self.Hbar**2 / self.Me
        )  # Bohr radius [A] 0.5291772
        self.Hartree = (
            self.Me * self.EV / 16 / pi**2 / self.Epsilon0**2 / self.Hbar**2
        )  # Hartree [eV] 27.211398
        self.Rydberg = self.Hartree / 2  # Rydberg [eV]

        self.THzToEv = self.PlanckConstant * 1e12  # [eV]
        self.KB = self.KB_J / self.EV  # [eV/K]
        self.THzToCm = 1.0e12 / (self.SpeedOfLight * 100)  # [cm^-1]
        self.CmToEv = self.THzToEv / self.THzToCm  # [eV]
        self.EVAngstromToGPa = self.EV * 1e21
        self.EvTokJmol = self.EV / 1000 * self.Avogadro  # [kJ/mol]

        self.DefaultToTHz = (
            sqrt(self.EV / self.AMU) / self.Angstrom / (2 * pi) / 1e12
        )  # [THz]


@dataclass(frozen=True)
class PhysicalUnits:
    """Immutable physical units for phonon calculations.

    Attributes
    ----------
    KB_J: float
        Boltzmann constant in J/K.
    PlanckConstant: float
        Planck constant in eV s.
    Avogadro: float
        Avogadro's number.
    SpeedOfLight: float
        Speed of light in m/s.
    AMU: float
        Atomic mass unit in kg.
    EV: float
        Electron volt in J.
    Me: float
        Electron mass in kg.

    THz: float
        Terahertz frequency in /s.
    Angstrom: float
        Angstrom in m.
    Newton: float
        Newton in kg m / s^2.
    Joule: float
        Joule in kg m^2 / s^2.
    Hbar: float
        Reduced Planck constant in eV s.
    Epsilon0: float
        Vacuum permittivity in C^2 / N m^2.
    Bohr: float
        Bohr radius in A.
    Hartree: float
        Hartree energy in eV.
    Rydberg: float
        Rydberg energy in eV.
    KB: float
        Boltzmann constant in eV/K.
    THzToEv: float
        Conversion factor from THz to eV.
    CmToEv: float
        Conversion factor from cm^-1 to eV.
    EVAngstromToGPa: float
        Conversion factor from eV/Angstrom to GPa.
    EvTokJmol: float
        Conversion factor from eV to kJ/mol.

    DefaultToTHz: float
        Default conversion factor to THz.

    """

    KB_J: float  # [J/K]
    PlanckConstant: float  # [eV s]
    Avogadro: float
    SpeedOfLight: float  # [m/s]
    AMU: float  # [kg]
    EV: float  # [J]
    Me: float  # [kg]
    THz: float  # [/s]
    Angstrom: float  # [m]
    Newton: float  # [kg m / s^2]
    Joule: float  # [kg m^2 / s^2]
    Hbar: float  # [eV s]
    Epsilon0: float  # [C^2 / N m^2]
    Bohr: float  # Bohr radius [A]
    Hartree: float  # Hartree [eV]
    Rydberg: float  # Rydberg [eV]
    THzToEv: float  # [eV]
    KB: float  # [eV/K]
    THzToCm: float  # [cm^-1]
    CmToEv: float  # [eV]
    EVAngstromToGPa: float
    EvTokJmol: float  # [kJ/mol]
    DefaultToTHz: float  # [THz]


# Global variable to store physical units
_physical_units: Optional[PhysicalUnits] = None


def set_physical_units(
    KB_J: float = 1.3806504e-23,  # [J/K]
    PlanckConstant: float = 4.13566733e-15,  # [eV s]
    Avogadro: float = 6.02214179e23,
    SpeedOfLight: float = 299792458,  # [m/s]
    AMU: float = 1.6605402e-27,  # [kg]
    EV: float = 1.60217733e-19,  # [J]
    Me: float = 9.10938215e-31,  # [kg],
):
    """Set physical units used globally.

    Default values are:

    KB_J : 1.3806504e-23 [J/K]
    PlanckConstant : 4.13566733e-15 [eV s]
    Avogadro : 6.02214179e+23
    SpeedOfLight : 299792458 [m/s]
    AMU : 1.6605402e-27 [kg]
    EV : 1.60217733e-19 [J]
    Me : 9.10938215e-31 [kg]
    THz : 1e12 [/s]
    Angstrom : 1e-10 [m]
    Newton : 1.0 [kg m / s^2]
    Joule : 1.0 [kg m^2 / s^2]
    Hbar : 6.582118985531608e-16 [eV s]
    Epsilon0 : 8.85418781762039e-12 [C^2 / N m^2]
    Bohr : 0.529177207423948 [A]
    Hartree : 27.211398230887998 [eV]
    Rydberg : 13.605699115443999 [eV]
    THzToEv : 0.00413566733 [eV]
    KB : 8.617338256808316e-05 [eV/K]
    THzToCm : 33.3564095198152 [cm^-1]
    CmToEv : 0.00012398418743309975 [eV]
    EVAngstromToGPa : 160.21773299999998 [GPa]
    EvTokJmol : 96.4853905398362 [kJ/mol]
    DefaultToTHz : 15.633302300230191 [THz]

    """
    global _physical_units
    physical_units = PhysicalUnitsGenerator(
        KB_J=KB_J,
        PlanckConstant=PlanckConstant,
        Avogadro=Avogadro,
        SpeedOfLight=SpeedOfLight,
        AMU=AMU,
        EV=EV,
        Me=Me,
    )
    _physical_units = PhysicalUnits(**asdict(physical_units))


def get_physical_units() -> PhysicalUnits:
    """Get physical units used globally."""
    return _physical_units


set_physical_units()



# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Thermal displacement plot (tdplot)
#
# Usage:
#   tdplot -i "1 2, 4 5" -o "td.pdf"
#
# The axis resolved thermal displacements are averaged by root mean
# square with the successive indices separated by ",". In this
# example, values at a temperature with the indices 1 and 2, 3 and 4
# are averaged by root mean square respectively as follwos:
#   sqrt( ( x_1 ** 2 + x_2 ** 2 + x_3 ** 2 + x_4 ** 2 ) / 4 ),
# and then they are ploted as a function of temperature.
#
# The definition is that indices correspond to those as follows:
# 1 2 3 : X Y Z of the 1st atom,
# 4 5 6 : X Y Z of the 2nd atom,
# ...

import sys

import matplotlib.pyplot as plt
import numpy as np

try:
    import yaml
except ImportError:
    print("You need to install python-yaml.")
    sys.exit(1)

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


def get_options():
    """Parse command-line options."""
    import argparse

    parser = argparse.ArgumentParser(description="Phonopy tdplot command-line-tool")
    parser.set_defaults(
        output_filename=None,
        factor=1.0,
        td_indices=None,
        show_legend=False,
        is_distance=False,
        is_gnuplot=False,
        ymax=None,
        ymin=None,
        tmax=None,
        tmin=None,
    )
    parser.add_argument(
        "-i", "--indices", dest="td_indices", help="Indices like 1 2, 3 4 5 6..."
    )
    parser.add_argument(
        "--factor",
        dest="factor",
        type=float,
        help="Conversion factor of energy unit to internal electronic energy",
    )
    parser.add_argument(
        "--ymax", dest="ymax", type=float, help="Maximum value of y axis"
    )
    parser.add_argument(
        "--ymin", dest="ymin", type=float, help="Minimum value of y axis"
    )
    parser.add_argument(
        "--tmax", dest="tmax", type=float, help="Maximum value of temperature"
    )
    parser.add_argument(
        "--tmin", dest="tmin", type=float, help="Minimum value of temperature"
    )
    parser.add_argument(
        "-o", "--output", dest="output_filename", help="Output filename"
    )
    parser.add_argument(
        "-l", "--legend", dest="show_legend", action="store_true", help="Show legend"
    )
    parser.add_argument(
        "--distance",
        dest="is_distance",
        action="store_true",
        help="Plot thermal distance",
    )
    parser.add_argument(
        "filename",
        nargs="*",
        help=(
            "Filename of phonon thermal displacement result "
            "(thermal_displacements.yaml)"
        ),
    )
    args = parser.parse_args()
    return args


def run():
    """Run phonopy-tdplot."""
    args = get_options()

    if not args.is_distance:
        if len(args.filename) == 0:
            filename = "thermal_displacements.yaml"
        else:
            filename = args.filename[0]
        yamldata = yaml.load(open(filename).read(), Loader=Loader)

        td = yamldata["thermal_displacements"]
        temperatures = [v["temperature"] for v in td]
        displacements = [v["displacements"] for v in td]
        displacements = np.array(displacements).reshape(len(temperatures), -1)
    else:
        if len(args.filename) == 0:
            filename = "thermal_distances.yaml"
        else:
            filename = args.filename[0]
        yamldata = yaml.load(open(filename).read(), Loader=Loader)

        td = yamldata["thermal_distances"]
        temperatures = [v["temperature"] for v in td]
        distances = [v["distance"] for v in td]
        for t, dists in zip(temperatures, distances):
            print(("%14.7f" * (len(dists) + 1)) % ((t,) + tuple(dists)))
            print("")
        print("")
        print("")
        sys.exit(0)

    # Set temperature range
    tmin_index = 0
    tmax_index = len(temperatures)
    if args.tmin is not None:
        for i, t in enumerate(temperatures):
            if t > args.tmin - (temperatures[1] - temperatures[0]) * 0.1:
                tmin_index = i
                break

    if args.tmax is not None:
        for i, t in enumerate(temperatures):
            if t > args.tmax + (temperatures[1] - temperatures[0]) * 0.1:
                tmax_index = i
                break

    # Extract indices
    set_of_indices = []
    if args.td_indices is None:
        set_of_indices.append(range(1, displacements.shape[1] + 1))
    else:
        for v in args.td_indices.split(","):
            set_of_indices.append([int(x) for x in v.split()])

    # Plot
    fig, ax = plt.subplots()
    ax.xaxis.set_ticks_position("both")
    ax.yaxis.set_ticks_position("both")
    ax.xaxis.set_tick_params(which="both", direction="in")
    ax.yaxis.set_tick_params(which="both", direction="in")

    plots = []
    for indices in set_of_indices:
        d = np.zeros(len(temperatures), dtype="double")
        for i in indices:
            d += displacements[:, i - 1]
        (curve,) = plt.plot(
            temperatures[tmin_index:tmax_index], d[tmin_index:tmax_index] / len(indices)
        )
        plots.append(curve)

    ax.set_xlim((0, temperatures[tmax_index - 1]))

    # Set y range
    if (args.ymin is not None) and (args.ymax is not None):
        plt.ylim(args.ymin, args.ymax)
    elif (args.ymin is None) and (args.ymax is not None):
        plt.ylim(ymax=args.ymax)
    elif (args.ymin is not None) and (args.ymax is None):
        plt.ylim(ymin=args.ymin)

    if args.show_legend:
        plt.legend(plots, set_of_indices, loc="upper left")

    # Set output device and show
    if args.output_filename is not None:
        plt.rcParams["pdf.fonttype"] = 42
        plt.rcParams["font.family"] = "serif"
        plt.savefig(args.output_filename)
    else:
        plt.show()



"""Tests for symfc interface."""

from typing import Optional

import pytest

from phonopy import Phonopy
from phonopy.interface.symfc import SymfcFCSolver


@pytest.mark.parametrize("cutoff", [None, {3: 5.0}])
def test_symfc_cutoff(ph_nacl: Phonopy, cutoff: Optional[dict]):
    """Test symfc interface with cutoff distance."""
    ph = ph_nacl
    symfc_solver = SymfcFCSolver(
        ph.supercell,
        ph.symmetry,
        options={"cutoff": cutoff},
        log_level=1,
    )
    symfc_solver.compute_basis_set(
        orders=[2, 3],
    )
    if cutoff is None:
        assert symfc_solver.basis_set[3].basis_set.shape == (786, 758)
    else:
        assert symfc_solver.basis_set[3].basis_set.shape == (80, 67)



# Copyright (C) 2025 Eugene Roginskii
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import argparse
from os import path

import numpy as np

from phonopy.cui import load

Debye2au = 2.54174


def get_cp2kver(fn):
    """Get CP2K version from the output file."""
    try:
        fh = open(fn, "r")
        for line in fh:
            if "CP2K| version string" in line:
                try:
                    ver = float(line.split("CP2K version")[1])
                    return int(ver)
                    break
                except ValueError:
                    print("Warning. Reading CP2K version failed. Suggest v8")
                    return 8
    except IOError:
        print("Warning. Reading CP2K version failed. Suggest v8 and above")
        return 8


def get_dipole(fn):
    """Get dipole moment from the output file."""
    dipole = []
    xyz = ["X=", "Y=", "Z="]
    try:
        fh = open(fn, "r")
        for line in fh:
            if "Dipole moment [Debye]" in line:
                break
        line = fh.readline()
        for i in range(3):
            dipole.append(float(line.split(xyz[i])[1].split()[0]) / Debye2au)
    except OSError as err:
        print("Error opening file %s: %s" % (fn, err))
    return np.array(dipole)


def get_epsilon_cp2k(fn, ucvol):
    """Get epsilon from the CP2K output file."""
    from numpy import pi

    epsilon = np.zeros(9)
    try:
        fh = open(fn, "r")
    except IOError:
        print("ERROR open output file %s for reading filed" % fn)
        return -1
    for line in fh:
        if "Polarizability tensor [a.u.]" in line:
            break
    # Components sequence in CP2K output
    # xx,yy,zz
    # xy,xz,yz
    # yx,zx,zy
    # Components sequence in epsilon array:
    #                     0    1    2    3    4    5    6    7    8
    #      9-components:  xx   xy   xz   yx   yy   yz   zx   zy   zz
    line = fh.readline()
    for i in range(3):
        epsilon[i * 4] = float(line.split()[i + 2]) / ucvol * 4 * pi + 1

    line = fh.readline()
    for i in range(2):
        epsilon[1 + i] = float(line.split()[i + 2]) / ucvol * 4 * pi
    epsilon[5] = float(line.split()[4]) / ucvol * 4 * pi

    line = fh.readline()
    epsilon[3] = float(line.split()[2]) / ucvol * 4 * pi
    for i in range(2):
        epsilon[6 + i] = float(line.split()[i + 3]) / ucvol * 4 * pi
    # Symmitrize
    epsilon[1] = (epsilon[1] + epsilon[3]) / 2
    epsilon[3] = epsilon[1]
    epsilon[2] = (epsilon[2] + epsilon[6]) / 2
    epsilon[6] = epsilon[2]
    epsilon[5] = (epsilon[5] + epsilon[7]) / 2
    epsilon[7] = epsilon[5]

    return epsilon


def get_epsilon_cp2kv6(fn, ucvol):
    """Get epsilon from the CP2K v6 output file."""
    from numpy import pi

    epsilon = np.zeros(9)
    print("cp2kv6")
    try:
        fh = open(fn, "r")
    except IOError:
        print("ERROR open output file %s for reading filed" % fn)
        return -1
    for line in fh:
        if "POLARIZABILITY TENSOR (atomic units)" in line:
            print(line)
            break

    line = fh.readline()
    for i in range(3):
        epsilon[i * 4] = float(line.split()[i + 1]) / ucvol * 4 * pi

    line = fh.readline()
    for i in range(2):
        epsilon[1 + i] = float(line.split()[i + 1]) / ucvol * 4 * pi
    epsilon[5] = float(line.split()[3]) / ucvol * 4 * pi

    line = fh.readline()
    epsilon[3] = float(line.split()[1]) / ucvol * 4 * pi
    for i in range(2):
        epsilon[6 + i] = float(line.split()[i + 2]) / ucvol * 4 * pi
    # Symmitrize
    epsilon[1] = (epsilon[1] + epsilon[3]) / 2
    epsilon[3] = epsilon[1]
    epsilon[2] = (epsilon[2] + epsilon[6]) / 2
    epsilon[6] = epsilon[2]
    epsilon[5] = (epsilon[5] + epsilon[7]) / 2
    epsilon[7] = epsilon[5]

    return epsilon


def is_minus_displacement(direction, site_symmetry):
    """Symmetrically check if minus displacement is necessary or not."""
    is_minus = True
    for r in site_symmetry:
        rot_direction = np.dot(direction, r.T)
        if (rot_direction + direction).any():
            continue
        else:
            is_minus = False
            break
    return is_minus


parser = argparse.ArgumentParser(description="Born charges file generator")

parser.add_argument(
    "-i",
    "--input",
    action="store",
    type=str,
    dest="eps_fn",
    help="CP2K output filename with LR calculations",
)
parser.add_argument(
    "-p",
    "--prefix",
    action="store",
    type=str,
    dest="pref",
    default="DISP",
    help="Prefix of directories with calculations",
)
parser.add_argument(
    "-m",
    "--dipolefn",
    action="store",
    type=str,
    dest="dipole_fn",
    default="force.out",
    help="CP2K output filename with dipole calculations",
)
parser.add_argument(
    "-o",
    "--out",
    action="store",
    type=str,
    dest="out_fn",
    default="BORN",
    help="Output filename (default: BORN)",
)

args = parser.parse_args()

b2a = 0.5291772109038
ph = load.load(phonopy_yaml="phonopy_disp.yaml")

basis = ph.supercell.cell
ucvol = ph.supercell.volume / (b2a**3)  # In a.u.

# Get epsilon
fn = path.join("%s_ideal" % args.pref, "polar.out")
if get_cp2kver(args.eps_fn) <= 6:
    epsilon = get_epsilon_cp2kv6(args.eps_fn, ucvol)
else:
    epsilon = get_epsilon_cp2k(args.eps_fn, ucvol)

dipoles = []
atlist = {}
for i, atom in enumerate(ph.dataset["first_atoms"]):
    wdir = "%s-%04d" % (args.pref, (i + 1))
    dipole = get_dipole(path.join(wdir, args.dipole_fn))
    dispmag = atom["displacement"].sum()
    dipoles.append([atom["number"], dispmag, dipole])
    if atom["number"] not in atlist.keys():
        # store index of displacement for each atom for positive and negative
        atlist.update({atom["number"]: [[-1, -1, -1], [-1, -1, -1]]})

for i, atom in enumerate(ph.dataset["first_atoms"]):
    ddir = np.where(atom["displacement"] != 0)[0][0]
    if atom["displacement"][ddir] > 0:
        atlist[atom["number"]][0][ddir] = i
    else:
        atlist[atom["number"]][1][ddir] = i

born = np.zeros(len(atlist) * 9).reshape(len(atlist), 3, 3)
for i, a in enumerate(atlist):
    for j in range(3):
        born[i][j] = (dipoles[atlist[a][0][j]][2] - dipoles[atlist[a][1][j]][2]) * b2a
        born[i][j] /= dipoles[atlist[a][0][j]][1] - dipoles[atlist[a][1][j]][1]
with open(args.out_fn, "w") as fh:
    fh.write("0.280028521\n")
    fh.write(" %12.9f " * 9 % tuple(epsilon))
    fh.write("\n ")
    for b in born:
        fh.write(" %12.9f " * 9 % tuple(b.flatten()))
        fh.write("\n ")



"""CP2K calculator interface."""

# vim: set fileencoding=utf-8 :
# Copyright (C) 2017-2019 Tiziano Müller
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function

import sys
from fractions import Fraction
from os import path

import numpy as np

from phonopy.file_IO import iter_collect_forces
from phonopy.interface.vasp import check_forces, get_drift_forces
from phonopy.structure.atoms import PhonopyAtoms, symbol_map


def cp2kver(forces_filename):
    tst2025 = ['FORCES| Atomic forces', 'FORCES| Total atomic force']
    is2025 = []
    havetoten = ''
    with open(forces_filename, 'r') as fd:
        for testline in tst2025:
            for line in fd:
                if testline in line:
                    is2025.append(True)
        fd.seek(0)
        for line in fd:
            if 'TOTEN' in line:
                havetoten = 'e'

    if len(is2025) > 0 and all(is2025):
        return '2025' + havetoten

    return '2024' + havetoten


def parse_set_of_forces(num_atoms, forces_filenames, verbose=True):
    if cp2kver(forces_filenames[0]) == '2025':
        return parse_set_of_forces2025(num_atoms,
                                       forces_filenames,
                                       verbose=verbose)
    elif cp2kver(forces_filenames[0]) == '2025e':
        return parse_set_of_energy_forces2025(num_atoms,
                                       forces_filenames,
                                       verbose=verbose)
    elif cp2kver(forces_filenames[0]) == '2024e':
        return parse_set_of_energy_forces2024(num_atoms,
                                       forces_filenames,
                                       verbose=verbose)
    else:
        return parse_set_of_forces2024(num_atoms,
                                       forces_filenames,
                                       verbose=verbose)


def parse_set_of_energy_forces2025(num_atoms, forces_filenames, verbose=True):
    """Parse forces from output files."""
    force_sets = []
    energy_sets = []

    for i, filename in enumerate(forces_filenames):
        if verbose:
            sys.stdout.write("%d. " % (i + 1))

        forces = iter_collect_forces(
            filename, num_atoms, "FORCES|   Atom     x", [2, 3, 4]
        )

        if not check_forces(forces, num_atoms, filename, verbose=verbose):
            return []  # if one file is invalid, the whole thing is broken

        drift_force = get_drift_forces(forces,
                                       filename=filename,
                                       verbose=verbose)
        force_sets.append(np.array(forces) - drift_force)

        esc = 0
        with open(filename, 'r') as fd:
            for line in fd:
                if 'TOTEN' in line:
                    esc = float(line.split('=')[1])
            energy_sets.append(esc)

    return {"forces": force_sets,
            "supercell_energies": energy_sets,
    }


def parse_set_of_forces2025(num_atoms, forces_filenames, verbose=True):
    """Parse forces from output files."""
    force_sets = []

    for i, filename in enumerate(forces_filenames):
        if verbose:
            sys.stdout.write("%d. " % (i + 1))

        forces = iter_collect_forces(
            filename, num_atoms, "FORCES|   Atom     x", [2, 3, 4]
        )

        if not check_forces(forces, num_atoms, filename, verbose=verbose):
            return []  # if one file is invalid, the whole thing is broken

        drift_force = get_drift_forces(forces,
                                       filename=filename,
                                       verbose=verbose)
        force_sets.append(np.array(forces) - drift_force)

    return force_sets


def parse_set_of_energy_forces2024(num_atoms, forces_filenames, verbose=True):
    """Parse forces from output files."""
    force_sets = []
    energy_sets = []

    for i, filename in enumerate(forces_filenames):
        if verbose:
            sys.stdout.write("%d. " % (i + 1))

        forces = iter_collect_forces(
            filename, num_atoms, "# Atom   Kind   Element", [3, 4, 5]
        )

        if not check_forces(forces, num_atoms, filename, verbose=verbose):
            return []  # if one file is invalid, the whole thing is broken

        drift_force = get_drift_forces(forces,
                                       filename=filename,
                                       verbose=verbose)
        force_sets.append(np.array(forces) - drift_force)

        esc = 0
        with open(filename, 'r') as fd:
            for line in fd:
                if 'TOTEN' in line:
                    esc = float(line.split('=')[1])
            energy_sets.append(esc)

    return {"forces": force_sets,
            "supercell_energies": energy_sets,
    }

def parse_set_of_forces2024(num_atoms, forces_filenames, verbose=True):
    """Parse forces from output files."""
    force_sets = []

    for i, filename in enumerate(forces_filenames):
        if verbose:
            sys.stdout.write("%d. " % (i + 1))

        forces = iter_collect_forces(
            filename, num_atoms, "# Atom   Kind   Element", [3, 4, 5]
        )

        if not check_forces(forces, num_atoms, filename, verbose=verbose):
            return []  # if one file is invalid, the whole thing is broken

        drift_force = get_drift_forces(forces,
                                       filename=filename,
                                       verbose=verbose)
        force_sets.append(np.array(forces) - drift_force)

    return force_sets


def read_cp2k(filename):
    """Read crystal structure."""
    from cp2k_input_tools.parser import CP2KInputParser

    with open(filename) as fhandle:
        parser = CP2KInputParser()
        tree = parser.parse(fhandle)

    try:
        subsys = tree["+force_eval"][0]["+subsys"]
        cp2k_cell = subsys["+cell"]
    except IndexError:
        raise RuntimeError(
            "could not find a FORCE_EVAL/SUBSYS/CELL section in the given "
            "CP2K input file"
        )

    if len(tree["+force_eval"]) > 1:
        raise NotImplementedError(
            "the given CP2K input file contains multiple FORCE_EVAL sections, "
            "which is not (yet) supported"
        )

    # CP2K can get its cell information in two ways:
    # - A, B, C: cell vectors
    # - ABC: scaling of cell vectors, ALPHA_BETA_GAMMA: angles between the cell vectors
    # We'll parse either of them, but only write A, B, C
    if "a" in cp2k_cell:
        # unit vectors given
        unit_cell = np.array(
            [
                cp2k_cell["a"],
                cp2k_cell["b"],
                cp2k_cell["c"],
            ]
        )
    elif "abc" in cp2k_cell:
        # length of unit vectors given
        if "alpha_beta_gamma" in cp2k_cell:
            # if we also have the angles, construct the cell

            alpha, beta, gamma = cp2k_cell.pop("alpha_beta_gamma")

            cos_alpha = np.cos(alpha * np.pi / 180.0)
            cos_beta = np.cos(beta * np.pi / 180.0)
            cos_gamma = np.cos(gamma * np.pi / 180.0)
            sin_gamma = np.sin(gamma * np.pi / 180.0)

            unit_cell = np.array(
                [
                    [1.0, 0.0, 0.0],
                    [cos_gamma, sin_gamma, 0.0],
                    [
                        cos_beta,
                        (cos_alpha - cos_gamma * cos_beta) / sin_gamma,
                        np.sqrt(
                            1.0
                            - cos_beta**2
                            - ((cos_alpha - cos_gamma * cos_beta) / sin_gamma) ** 2
                        ),
                    ],
                ]
            )
        else:
            unit_cell = np.eye(3)

        a, b, c = cp2k_cell.pop(
            "abc"
        )  # remove them from the tree since we pass it along

        unit_cell[0, :] *= a
        unit_cell[1, :] *= b
        unit_cell[2, :] *= c

    if "+cell_ref" in cp2k_cell:
        print("WARNING: the &CELL_REF section must be manually adjusted")

    cp2k_coord = subsys["+coord"]

    numbers = []
    positions = []

    for coordline in cp2k_coord["*"]:
        # coordinates are a series of strings according to the CP2K schema
        fields = coordline.split()
        numbers += [symbol_map[fields[0]]]
        # positions can also be fractions
        positions += [[float(Fraction(f)) for f in fields[1:4]]]

    if cp2k_coord.get("scaled", False):  # the keyword can be unavailable, true or false
        return (
            PhonopyAtoms(numbers=numbers, cell=unit_cell, scaled_positions=positions),
            tree,
        )
    else:
        return (
            PhonopyAtoms(numbers=numbers, cell=unit_cell, positions=positions),
            tree,
        )


def write_supercells_with_displacements(
    supercell,
    cells_with_displacements,
    ids,
    optional_structure_info,
    pre_filename="supercell",
    width=3,
):
    """Write supercells with displacements to files."""
    orig_fname, tree = optional_structure_info

    fbase, fext = path.splitext(orig_fname)
    pbase = tree["+global"]["project_name"]

    supercell_ref_name = "{}-supercell{}".format(fbase, fext)
    with open(supercell_ref_name, "w") as fhandle:
        fhandle.write(
            """\
# Generated by Phonopy, based on {fname}
# Original configuration with the generated supercell for comparison
""".format(
                fname=orig_fname
            )
        )
        write_cp2k(fhandle, "{}-{}".format(pbase, pre_filename), supercell, tree)

    for i, cell in zip(ids, cells_with_displacements):
        suffix = "{pre_filename}-{0:0{width}}".format(
            i, pre_filename=pre_filename, width=width
        )

        with open("{}-{}{}".format(fbase, suffix, fext), "w") as fhandle:
            fhandle.write(
                """\
# Generated by Phonopy, based on {fname}
# Merged configuration with displacements
""".format(
                    fname=orig_fname
                )
            )
            write_cp2k(fhandle, "{}-{}".format(pbase, suffix), cell, tree)


def write_cp2k_by_filename(filename, cell, tree, header=None):
    """Wrap write_cp2k to write an arbitrary unit cell.

    Note
    ----
    This method is written by Atsushi Togo to use at
        phonopy.interface.calculator.write_crystal_structure.
    I am not a user of cp2k and this method can be wrongly written.
    Please rewrite if somebody who knows well about cp2k input.

    """
    fbase, fext = path.splitext(filename)
    pbase = tree["+global"]["project_name"]
    project_name = "{}-{}".format(pbase, fbase)

    if header is None:
        _header = "# Generated by Phonopy\n"
    else:
        _header = header

    with open(filename, "w") as w:
        w.write(_header)
        write_cp2k(w, project_name, cell, tree)


def write_cp2k(fhandle, project_name, atoms, tree):
    """Merge the new the atoms structure with the configuration tree to a new CP2K input file.

    :param fhandle: open file handle to which the routine will write to
    :param project_name: the project name to use (CP2K uses that as prefix for generated files)i
    :param atoms: the Atoms objects to use
    :param tree: the configuration tree as returned from CP2KInputParser
    """
    from cp2k_input_tools.generator import CP2KInputGenerator

    generator = CP2KInputGenerator()

    tree["+global"]["run_type"] = "ENERGY_FORCE"
    tree["+global"]["project_name"] = project_name

    force_eval = tree["+force_eval"][0]
    subsys = force_eval["+subsys"]

    # if the original input contained scaled positions, continue with scaled positions
    if subsys["+coord"].get("scaled", False):
        cp2k_coord = {
            "scaled": True,
            "*": [
                "{sym} {x} {y} {z}".format(sym=sym, x=coord[0], y=coord[1], z=coord[2])
                for sym, coord in zip(
                    atoms.get_chemical_symbols(), atoms.get_scaled_positions()
                )
            ],
        }
    # ... otherwise use absolute positions
    else:
        cp2k_coord = {
            "*": [
                "{sym} {x} {y} {z}".format(sym=sym, x=coord[0], y=coord[1], z=coord[2])
                for sym, coord in zip(
                    atoms.get_chemical_symbols(), atoms.get_positions()
                )
            ],
        }

    subsys["+cell"]["a"] = list(atoms.get_cell()[0])
    subsys["+cell"]["b"] = list(atoms.get_cell()[1])
    subsys["+cell"]["c"] = list(atoms.get_cell()[2])
    subsys["+cell"]["periodic"] = "XYZ"  # anything else does not make much sense

    subsys["+coord"] = cp2k_coord  # overwriting the coordinates

    if "+print" not in force_eval:
        force_eval["+print"] = {}
    if "+forces" not in force_eval["+print"]:
        force_eval["+print"]["+forces"] = {}
    force_eval["+print"]["+forces"][
        "filename"
    ] = "forces"  # uses the project name as base with 'forces' as suffix

    for line in generator.line_iter(tree):
        fhandle.write("{line}\n".format(line=line))
